/*  Copyright 2004-2012, Unpublished Work of Technologic Systems
*  All Rights Reserved.
*
*  THIS WORK IS AN UNPUBLISHED WORK AND CONTAINS CONFIDENTIAL,
*  PROPRIETARY AND TRADE SECRET INFORMATION OF TECHNOLOGIC SYSTEMS.
*  ACCESS TO THIS WORK IS RESTRICTED TO (I) TECHNOLOGIC SYSTEMS EMPLOYEES
*  WHO HAVE A NEED TO KNOW TO PERFORM TASKS WITHIN THE SCOPE OF THEIR
*  ASSIGNMENTS  AND (II) ENTITIES OTHER THAN TECHNOLOGIC SYSTEMS WHO
*  HAVE ENTERED INTO  APPROPRIATE LICENSE AGREEMENTS.  NO PART OF THIS
*  WORK MAY BE USED, PRACTICED, PERFORMED, COPIED, DISTRIBUTED, REVISED,
*  MODIFIED, TRANSLATED, ABRIDGED, CONDENSED, EXPANDED, COLLECTED,
*  COMPILED,LINKED,RECAST, TRANSFORMED, ADAPTED IN ANY FORM OR BY ANY
*  MEANS,MANUAL, MECHANICAL, CHEMICAL, ELECTRICAL, ELECTRONIC, OPTICAL,
*  BIOLOGICAL, OR OTHERWISE WITHOUT THE PRIOR WRITTEN PERMISSION AND
*  CONSENT OF TECHNOLOGIC SYSTEMS . ANY USE OR EXPLOITATION OF THIS WORK
*  WITHOUT THE PRIOR WRITTEN CONSENT OF TECHNOLOGIC SYSTEMS  COULD
*  SUBJECT THE PERPETRATOR TO CRIMINAL AND CIVIL LIABILITY.
*/
/* To compile tshwctl, use the appropriate cross compiler and run the
* command:
*
*  gcc -fno-tree-cselim -Wall -O0 -mcpu=arm9 -o tshwctl tshwctl.c ispvm.c \
    ispspi.c
*
* You need ispvm.c and vmopcode.h in the same directory as tshwctl to
* compile.
*
* On uclibc based initrd's, the following additional gcc options are
* necessary: -Wl,--rpath,/slib -Wl,-dynamic-linker,/slib/ld-uClibc.so.0
*
* If you want --loadfpga to work with .jed files, the executable "jed2vme"
* should be in the $PATH during runtime.
*/

const char copyright[] = "Copyright (c) Technologic Systems - " __DATE__ ;

#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <getopt.h>
#include <libgen.h>
#include <limits.h>
#include <linux/i2c-dev.h>
#include <linux/unistd.h>
#include <netdb.h>
#include <net/if.h>
#include <sched.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/ipc.h>
#include <sys/mman.h>
#include <sys/sem.h>
#include <sys/shm.h>
#include <sys/socket.h>
#include <sys/select.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/times.h>
#include <sys/timex.h>
#include <sys/types.h>
#include <time.h>
#include <time.h>
#include <unistd.h>

#ifndef TEMP_FAILURE_RETRY
# define TEMP_FAILURE_RETRY(expression) \
  (__extension__                                                              \
    ({ long int __result;                                                     \
      do __result = (long int) (expression);                                 \
      while (__result == -1L && errno == EINTR);                             \
      __result; }))
#endif

#define DIOMASK 0x33ffffffffffc3dfULL

volatile unsigned int *mxspiregs, *mxgpioregs, *mxtwiregs;
volatile unsigned int *mxmiscregs, *mxtimerregs;
volatile unsigned int *mxgpioregs;
unsigned int gpio_state[4];
int long_tagmem = 0;
static int nbuslocked = 0;

extern signed char ispVM(char *);
extern signed char ispSPI(char *);
void nbuslock(void);
void nbusunlock(void);
void nbuspreempt(void);
void nbus_poke16(unsigned char, unsigned short);
unsigned short nbus_peek16(unsigned char);
//int i2c_write(int chip, int addr, int count, char *buf);

volatile unsigned int *fe_base;

struct vtu {
	int v, vid;
	unsigned char tags[7];
	unsigned char forceVID[7];
};                                

static unsigned short model;
static int cpu_port;
#define FE_BASE_PHYSICAL 0xC0800000                          
#define SMI_RDVALID     (1 << 27)                            
#define SMI_BUSY        (1 << 28)                            
                                                             
#define GLOBAL_REGS_1   0x1f                                 
#define GLOBAL_REGS_2   0x17                                 
                                                             
#define VTU_OPS_REGISTER  0x05  /* offset in GLOBAL_REGS_1 */
#define VTU_OP_FLUSHALL          (1 << 12)                   
#define VTU_OP_LOAD              (3 << 12)                   
#define VTU_OP_GET_NEXT          (4 << 12)                   
#define VTU_OP_GET_CLR_VIOLATION (7 << 12)                   
                                                             
#define VTU_VID_REG     0x06  /* offset in GLOBAL_REGS_1 */  
#define FPGA_SYSCON_PHYSICAL    0x80004000
#define DIO27  1                          
#define DIO42  (1<<15)                    
#define MDIO   DIO27                      
#define MDC    DIO42 

static unsigned int has_switch(void) __attribute__((pure));
static unsigned int has_switch(void) {
	switch(model) {
	  case 0x4600:
	  case 0x7680:
		return 1;
		break;
	  default:
		return 0;
		break;
	}
}

static unsigned int is_7400(void) __attribute__((pure));
static unsigned int is_7400(void) {
	FILE *cpuinfo;
	char buf[512];
	char *ptr = NULL;;
	static unsigned int checked = 0;
	static unsigned int is7400 = 0;

	if(!checked) {
		bzero(buf, sizeof(buf));
		cpuinfo = fopen("/proc/device-tree/model", "r");
		if(cpuinfo != NULL) {
			fread(buf, 1, sizeof(buf)-1, cpuinfo);
			fclose(cpuinfo);
			if((ptr = strstr(buf, "TS-")) != NULL) {
				model = strtoul(ptr+3, NULL, 16);
			} else {
				fprintf(stderr, 
				  "device-tree does not contain model\n");
				exit(4);
			}

			if(model == 0x7400) is7400 = 1;
			if(model == 0x7670) is7400 = 1;
			if(model == 0x7680) is7400 = 1;
		} else {
			cpuinfo = fopen("/proc/cpuinfo", "r");
			if(cpuinfo == NULL) {
				perror("/proc/cpuinfo");
				exit(4);
			}
			fread(buf, 1, sizeof(buf)-1, cpuinfo);
			fclose(cpuinfo);
			if(strstr(buf, "TS-7400")) is7400 = 1;

			if(is7400) {
				cpuinfo = fopen("/dev/tsmodel", "r");
				if(cpuinfo == NULL) {
					perror("/dev/tsmodel");
					exit(4);
				}
				bzero(buf, sizeof(buf));
				fread(buf, 1, sizeof(buf)-1, cpuinfo);
				fclose(cpuinfo);
				model = strtoul(buf, NULL, 16);
			}
		}
		checked = 1;

	}
	return is7400;
}

/*XXX: DO NOT USE THESE ROUTINES IN A CUSTOMER APPLICATION UNLESS YOU FULLY
* UNDERSTAND HOW THEY ARE USED AND WHAT THEY DO! See nbus.c from our FTP site
* for the API intended for customer use!
*/
void poke16(unsigned char adr, unsigned short dat)
{
	int x;
	
	mxgpioregs[0x708/4] = (1 << 16) | (1 << 25) | (0xff); 
	mxgpioregs[0x704/4] = (1 << 24) | (1 << 26) | adr ;
	mxgpioregs[0x704/4] = (1 << 25);

	for(x = 1; x >= 0; x--) {
		mxgpioregs[0x708/4] = (1 << 26) | (1 << 25) | (0xff);
		mxgpioregs[0x704/4] = (unsigned char)(dat >> (x*8));
		mxgpioregs[0x704/4] = (1 << 25);
	}
	mxgpioregs[0x704/4] = (1 << 16);

	while(mxgpioregs[0x900/4] & (1 << 21)) {
		mxgpioregs[0x708/4] = (1 << 16);
		mxgpioregs[0x704/4] = (1 << 16);
	}
}

unsigned short peek16(unsigned char adr)
{
	int x;
	unsigned short ret;

	mxgpioregs[0x708/4] = (1 << 16) | (1 << 25) | (1 << 24) | (0xff);
	mxgpioregs[0x704/4] = ((1 << 26) | adr);
	mxgpioregs[0x704/4] = (1 << 25);
	mxgpioregs[0xb08/4] = 0xff;

	do {
		ret = 0;
		for(x = 1; x >= 0; x--) {
			mxgpioregs[0x708/4] = (1 << 26) | (1 << 25) | (0xff);
			mxgpioregs[0x704/4] = (1 << 25);
			ret |= ((mxgpioregs[0x900/4] & 0xff) << (x*8));
		}
		mxgpioregs[0x704/4] = (1 << 16);
		mxgpioregs[0x708/4] = (1 << 16);
	} while(mxgpioregs[0x900/4] & (1 << 21));

	mxgpioregs[0x704/4] = (1 << 16);
	mxgpioregs[0xb04/4] = 0xff;

	return ret;
}

static int semid = -1;
void nbuslock(void)
{
	int r;
	struct sembuf sop;

	if (semid == -1) {
		key_t semkey;
		semkey = 0x75000000;
		semid = semget(semkey, 1, IPC_CREAT|IPC_EXCL|0777);
		if (semid != -1) {
			sop.sem_num = 0;
			sop.sem_op = 1;
			sop.sem_flg = 0;
			r = semop(semid, &sop, 1);
			assert (r != -1);
		} else semid = semget(semkey, 1, 0777);
		assert (semid != -1);
	}
	sop.sem_num = 0;
	sop.sem_op = -1;
	sop.sem_flg = SEM_UNDO;


	/*Wrapper added to retry in case of EINTR*/
	r = TEMP_FAILURE_RETRY(semop(semid, &sop, 1));
	assert(r == 0);

	nbuslocked = 1;
}

void nbusunlock(void) 
{
	struct sembuf sop = { 0, 1, SEM_UNDO};
	int r;
	if(nbuslocked) {
		r = semop(semid, &sop, 1);
		assert(r == 0);
		nbuslocked = 0;
	}
}

void nbuspreempt(void)
{
	int r;
	r = semctl(semid, 0, GETNCNT);
	assert(r != -1);
	if(r) {
		nbusunlock();
		sched_yield();
		nbuslock();
	}
}

void sspi_cmd(unsigned int cmd) {
	unsigned int i;
	unsigned int s = peek16(0x6) & 0xfff0;
	
	// pulse CS#
	poke16(0x6, s | 0x2);
	poke16(0x6, s | 0x0);

	for (i = 0; i < 32; i++, cmd <<= 1) {
		if (cmd & 0x80) {
			poke16(0x6, s | 0x4);
			poke16(0x6, s | 0xc);
		} else {
			poke16(0x6, s | 0x0);
			poke16(0x6, s | 0x8);
		}
	}
}


int read_tagmem(unsigned int *tagmem) {
	int i, j, len;
	unsigned int ret;
	unsigned int s = peek16(0x6) & 0xfff0;

	sspi_cmd(0xac); // X_PROGRAM_EN
	sspi_cmd(0x4e); // READ_TAG

	if(long_tagmem) len = 24;
	else len = 20;
	
	for (j = 0; j < len; j++) {
		for (ret = 0x0, i = 0; i < 32; i++) {
			poke16(0x6, s | 0x0);
			poke16(0x6, s | 0x8);
			ret = ret << 1 | (peek16(0x6) & 0x1);
		}
		tagmem[j] = ret;
	}
	
	sspi_cmd(0x78); // PROGRAM_DIS
	
	poke16(0x6, s | 0x2);
	return 1;
}


int write_tagmem(unsigned int *tagmem) {
	int i, j, len;
	unsigned int s = peek16(0x6) & 0xfff0;
	
	sspi_cmd(0xac); // X_PROGRAM_EN
	sspi_cmd(0x8e); // WRITE_TAG

	if(long_tagmem) len = 24;
	else len = 20;
	
	for (j = 0; j < len; j++) {
		unsigned int x = tagmem[j];
		for (i = 0; i < 32; i++, x <<= 1) {
			if (x & 0x80000000UL) {
				poke16(0x6, s | 0x4);
				poke16(0x6, s | 0xc);
			} else {
				poke16(0x6, s | 0x0);
				poke16(0x6, s | 0x8);
			}
			if (!long_tagmem && i == 23 && j == 19) break;
		}
	}
	
	for (i = 0; i < 8; i++) {
		poke16(0x6, s | 0x2);
		poke16(0x6, s | 0xa);
	}
	poke16(0x6, s | 0x2);
	nbusunlock();
	usleep(25000);
	nbuslock();
	return 1;
}


int erase_tagmem() {
	int i;
	unsigned int s = peek16(0x6) & 0xfff0;

	sspi_cmd(0xac); // X_PROGRAM_EN
	sspi_cmd(0xe); // ERASE_TAG

	for (i = 0; i < 8; i++) {
		poke16(0x6, s | 0x2);
		poke16(0x6, s | 0xa);
	}
	poke16(0x6, s | 0x2);
	nbusunlock();
	usleep(1000000);
	nbuslock();
	return 1;
}

#define e2_read(r, b, c) i2c_op((0x50|((r&0x300)>>8)), b, r, 0x6, c)
#define e2_write(r, b, c) i2c_op((0x50|((r&0x300)>>8)), b, r, 0x7, c)
#define fpga_read(reg, buf, count) i2c_op(0x28, buf, reg, 0x4, count)
#define fpga_write(reg, buf, count) i2c_op(0x28, buf, reg, 0x5, count)
#define rtc_read(reg, buf, count) i2c_op(0x6f, buf, reg, 0x2, count)
#define rtc_write(reg, buf, count) i2c_op(0x6f, buf, reg, 0x3, count)
#define nvram_read(reg, buf, count) i2c_op(0x57, buf, reg, 0x2, count)
#define nvram_write(reg, buf, count) i2c_op(0x57, buf, reg, 0x3, count)
#define m0_read(buf,count) i2c_op(0x78, buf, 0, 0, count)
#define m0_write(buf, count) i2c_op(0x78, buf, 0, 0x1, count)

static const char *i2c_bus_imx,  /* the cortex, rtc are on this bus */
	*i2c_bus_gpio;		 /* and the fpga, eeprom are on this one */

int i2c_op(char adr, void *buf, short reg, short mode, char count)
{
	int i2cfd;
	int i;
	char adrbuf[3];
	char namebuf[16];

	if (i2c_bus_gpio == NULL || i2c_bus_imx == NULL) {
		FILE *name = fopen("/sys/class/i2c-dev/i2c-0/name", "r");
		if (name == NULL ||
		  fgets(namebuf, sizeof(namebuf), name) == NULL) {
			perror("/sys/class/i2c-dev/i2c-0/name");
			exit(4);
		} else {
			fclose(name);
			if (strstr(namebuf, "gpio")) {
				i2c_bus_gpio = "/dev/i2c-0";
				i2c_bus_imx = "/dev/i2c-1";
			} else if (strstr(namebuf, "MXS")) {
				i2c_bus_imx = "/dev/i2c-0";
				i2c_bus_gpio = "/dev/i2c-1";
			}
		}
	}
	if(mode >= 0x4 && mode <= 0x7) {
		i2cfd = open(i2c_bus_imx, O_RDWR);
		adrbuf[0] = ((reg >> 8) & 0xFF);
		adrbuf[1] = (reg & 0xFF);
	} else {
		i2cfd = open(i2c_bus_imx, O_RDWR);
	}

	assert(i2cfd != -1);

	ioctl(i2cfd, I2C_SLAVE_FORCE, adr);

	switch(mode) {
	  case 0x0: //adr, count
		i = read(i2cfd, buf, count);
		break;
	  case 0x1: //adr, count
		i = write(i2cfd, buf, count);
		break;
	  case 0x2: //adr, reg, count
	  case 0x6:
		for(i = 0; i < count; i++) {
			*(char *)(buf+i) = i2c_smbus_read_byte_data(i2cfd,
			  (unsigned char)(reg+i));
		}
		break;
	  case 0x3: //adr, reg, count
	  case 0x7:
		for(i = 0; i < count; i++) {
			i2c_smbus_write_byte_data(i2cfd, (unsigned char)(reg+i),
	        	  *(char *)(buf+i));
		}
		break;
	  case 0x4:
		write(i2cfd, &adrbuf , 2); //Write 2 byte reg address
		i = read(i2cfd, buf, count); //read back count bytes
		break;
	  case 0x5: /*Linux lacks a good way to handle this */
		for(i = 0; i < count; i++) {
			adrbuf[2] = *(char *)(buf+i);
			write(i2cfd, &adrbuf, 3);
			adrbuf[0] = ((++reg >> 8) & 0xFF);
			adrbuf[1] = (reg & 0xFF);
		}
		break;

	}

	close(i2cfd);

	return i;
}

#define SBC_MDIO_RD  ((mxgpioregs[0x940/4] >> 1) & 0x1)
#define SBC_MDC_RD  ((mxgpioregs[0x940/4] >> 0) & 0x1)
#define SBC_MDIO_OUT  mxgpioregs[0xb44/4] = 0x2
#define SBC_MDIO_IN  mxgpioregs[0xb48/4] = 0x2
#define SBC_MDC_OUT  mxgpioregs[0xb44/4] = 0x1
#define SBC_MDC_IN  mxgpioregs[0xb48/4] = 0x1
#define SBC_MDIO_HI  mxgpioregs[0x744/4] = 0x2
#define SBC_MDIO_LO  mxgpioregs[0x748/4] = 0x2
#define SBC_MDC_HI  mxgpioregs[0x744/4] = 0x1
#define SBC_MDC_LO  mxgpioregs[0x748/4] = 0x1

#define BB_MDIO_RD  ((peek16(0xe) >> 14) & 0x1)
#define BB_MDC_RD  ((peek16(0xe) >> 13) & 0x1)
#define BB_MDIO_OUT  poke16(0x28, (0x80 | 30))
#define BB_MDIO_IN  poke16(0x28, 30)
#define BB_MDC_OUT  poke16(0x28, (0x80 | 29))
#define BB_MDC_IN  poke16(0x28, 29)
#define BB_MDIO_HI  poke16(0x26, (0x80 | 30))
#define BB_MDIO_LO  poke16(0x26, 30)
#define BB_MDC_HI  poke16(0x26, (0x80 | 29))
#define BB_MDC_LO poke16(0x26, 29)

int opt_ethbb = 0;
static int inline MDIO_RD(void) {
	if(!opt_ethbb) return SBC_MDIO_RD;
	else return BB_MDIO_RD;
}
static int inline MDC_RD(void) {
	if(!opt_ethbb) return SBC_MDC_RD;
	else return BB_MDC_RD;
}
static void inline MDIO_OUT(void) {
	if(!opt_ethbb) SBC_MDIO_OUT;
	else BB_MDIO_OUT;
}
static void inline MDIO_IN(void) {
	if(!opt_ethbb) SBC_MDIO_IN;
	else BB_MDIO_IN;
}
static void inline MDC_OUT(void) {
	if(!opt_ethbb) SBC_MDC_OUT;
	else BB_MDC_OUT;
}
static void inline MDC_IN(void) {
	if(!opt_ethbb) SBC_MDC_IN;
	else SBC_MDC_IN;
}
static void inline MDIO_HI(void) {
	if(!opt_ethbb) SBC_MDIO_HI;
	else BB_MDIO_HI;
}
static void inline MDIO_LO(void) {
	if(!opt_ethbb) SBC_MDIO_LO;
	else BB_MDIO_LO;
}
static void inline MDC_HI(void) {
	if(!opt_ethbb) SBC_MDC_HI;
	else BB_MDC_HI;
}
static void inline MDC_LO(void) {
	if(!opt_ethbb) SBC_MDC_LO;
	else BB_MDC_LO;
}

static int phy_write(unsigned long phy, unsigned long reg, unsigned short data) {
	int x;
	unsigned int b;

	if(opt_ethbb) nbuslock();

	MDIO_OUT();
	MDC_OUT();

	MDC_LO();
	MDIO_HI();

	/* preamble, toggle clock high then low */
	for(x=0; x < 32; x++) {
		MDC_HI();
		MDC_LO();
	}
	/* preamble ends with MDIO high and MDC low */

	/* start bit followed by write bit */
	for(x=0; x < 2; x++) {
		MDIO_LO();
		MDC_HI();
		MDC_LO();
		MDIO_HI();
		MDC_HI();
		MDC_LO();
	}
	/* ends with MDIO high and MDC low */

	/* send the PHY address, starting with MSB */
	b = 0x10;
	for(x=0; x < 5; x++) {
		if (phy & b)
		  {MDIO_HI();}
		else
		  {MDIO_LO();}

		MDC_HI();
		MDC_LO();

		b >>= 1;
	}
	/* ends with MDC low, MDIO indeterminate */

	/* send the register address, starting with MSB */
	b = 0x10;
	for(x=0; x < 5; x++) {
		if (reg & b)
		  {MDIO_HI();}
		else
		  {MDIO_LO();}

		MDC_HI();
		MDC_LO();

		b >>= 1;
	}

	/* ends with MDC low, MDIO indeterminate */

	/* clock a one, then clock a zero */
	MDIO_HI();
	MDC_HI();
	MDC_LO();

	MDIO_LO();
	MDC_HI();
	MDC_LO();

	/* send the data, starting with MSB */
	b = 0x8000;
	for(x=0; x < 16; x++) {
		if (data & b)
		 { MDIO_HI();}
		else
		  {MDIO_LO();}

		MDC_HI();
		MDC_LO();

		b >>= 1;
	}

	MDIO_IN();
	MDC_IN();

	if(opt_ethbb) nbusunlock();
	return 0;
}

static int phy_read(unsigned long phy, unsigned long reg,
  volatile unsigned short *data) {
	int x, d;
	unsigned int a,b;

	d = 0;

	if(opt_ethbb) nbuslock();
	MDC_OUT();
	MDIO_OUT();

	MDC_LO();
	MDIO_HI();

	/* preamble, toggle clock high then low */
	for(x=0; x < 32; x++) {
		MDC_HI();
		MDC_LO();
	}
	/* preamble ends with MDIO high and MDC low */

	/* start bit */
	MDIO_LO(); MDC_HI(); MDC_LO();
	MDIO_HI(); MDC_HI(); MDC_LO();
	MDC_HI();  MDC_LO();
	MDIO_LO(); MDC_HI(); MDC_LO();

	/* send the PHY address, starting with MSB */
	b = 0x10;
	for(x=0; x < 5; x++) {
		if (phy & b)
		  {MDIO_HI();}
		else
		  {MDIO_LO();}

		MDC_HI();
		MDC_LO();

		b >>= 1;
		}
		/* ends with MDC low, MDIO indeterminate */

	/* send the register address, starting with MSB */
	b = 0x10;
	for(x=0; x < 5; x++) {
		if (reg & b)
		  {MDIO_HI();}
		else
		  {MDIO_LO();}

		MDC_HI();
		MDC_LO();

		b >>= 1;
	}

	MDIO_IN();
	/* ends with MDC low, MDIO indeterminate */

	/* another clock */
	MDC_HI();
	MDC_LO();

	/* read the data, starting with MSB */
	b = 0x8000;
	for(x=0; x < 16; x++) {
		MDC_HI();
		a = MDIO_RD();

		if (a & 1)
		d |= b;

		MDC_LO();

		b >>= 1;
	}

	MDIO_IN();
	MDC_IN();

	*data = d;
	if(opt_ethbb) nbusunlock();
	return 0;
}

static uint16_t vtu_readwait(unsigned long reg) {
	volatile unsigned short x;
	do {
		if(phy_read(GLOBAL_REGS_1, reg, &x) < 0) 
		  fprintf(stderr, "VTU_Read Timeout to 0x%lX\n", reg);
	} while (x & (1 << 15));

	return x;
}

static const char *MemberTagString(int tag) {
	switch(tag) {
		case 0: return "unmodified";
		case 1: return "untagged";
		case 2: return "tagged";
		default: return "unknown";
	}
}

inline void switch_errata_3_1(unsigned int n) {  
   volatile unsigned short smicmd, x;
   
   /* Write 0x0003 to PHY register 13 */
   do {
      phy_read(0x17, 0x18, &x); 
   } while (x & (1 << 15));
   smicmd = 0x960D | n;
	phy_write(0x17, 0x19, 0x0003);   /* smi data */		
	phy_write(0x17, 0x18, smicmd);
	
	/* Write 0x0000 to PHY register 14 */
	do {
	   phy_read(0x17, 0x18, &x); 
   } while (x & (1 << 15));
	smicmd = 0x960E | n;
	phy_write(0x17, 0x19, 0);   /* smi data */		
	phy_write(0x17, 0x18, smicmd);
	
	/* Write 0x4003 to PHY register 13 */
	do {
	   phy_read(0x17, 0x18, &x); 
   } while (x & (1 << 15));
   smicmd = 0x960D | n;
	phy_write(0x17, 0x19, 0x4003);   /* smi data */		
	phy_write(0x17, 0x18, smicmd);
	
	/* Write 0x0000 to PHY register 14 */
	do {
	   phy_read(0x17, 0x18, &x); 
   } while (x & (1 << 15));
	smicmd = 0x960E | n;
	phy_write(0x17, 0x19, 0);   /* smi data */		
	phy_write(0x17, 0x18, smicmd);
}

inline void switch_enphy(unsigned long phy) {
   volatile unsigned short x;
   do {
	   phy_read(0x17, 0x18, &x); 
   } while (x & (1 << 15));
	phy_write(0x17, 0x19, 0xb300);
	phy_write(0x17, 0x18, phy);
}

inline void switch_enflood(unsigned long port) {
	volatile unsigned short data;
	phy_read(port, 0x04, &data);
	phy_write(port, 0x04, data | 0xf);
}

inline void switch_enbcastflood(unsigned long port) {
	phy_write(port, 0x04, 0x7f);
}

inline void switch_forcelink(unsigned long port) {
	volatile unsigned short data;
	phy_read(port, 0x01, &data);
	phy_write(port, 0x01, data | 0x30);
}

static int vtu_add_entry(struct vtu *entry) {
	int i, j, p;
	volatile unsigned short x, r7, r8;
	unsigned short r6[7];

	while(1) {
		if (phy_read(GLOBAL_REGS_1, VTU_OPS_REGISTER, &x) < 0) {
			fprintf(stderr, "Timeout %s %d\n", __FILE__, __LINE__);
			return -1;
		}

		if (! (x & (1 << 15)))  /* wait for VB=0 in Global Reg 0x05 */
		break;
	}

	phy_write(GLOBAL_REGS_1, VTU_VID_REG, (1 << 12) | entry->vid);

	r7 = entry->tags[0] | (entry->tags[1] << 4) | (entry->tags[2] << 8) |
	  (entry->tags[3] << 12);
	r8 = entry->tags[4] | (entry->tags[5] << 4) | (entry->tags[6] << 8);

	phy_write(GLOBAL_REGS_1, 0x07, r7);
	phy_write(GLOBAL_REGS_1, 0x08, r8);
	phy_write(GLOBAL_REGS_1, VTU_OPS_REGISTER, 0xb000);   /* start the load */

	/* Next, we take care of the VLAN map, which is a per-port
	bitfield at offset 6 */

	while(1) {
		if (phy_read(GLOBAL_REGS_1, VTU_OPS_REGISTER, &x) < 0) {
			fprintf(stderr, "Timeout %s %d\n", __FILE__, __LINE__);
			return -1;
		}

		if (! (x & (1 << 15)))  /* wait for VB=0 in Global Reg 0x05 */
		break;
	}

	phy_write(GLOBAL_REGS_1, VTU_VID_REG, 0xFFF);

	i=j=0;
	memset(&r6, 0, sizeof(r6));

	while(1) {
		x = 0;
		phy_write(GLOBAL_REGS_1, VTU_OPS_REGISTER, (1 << 15) | VTU_OP_GET_NEXT);

		while(1) {
			if (phy_read(GLOBAL_REGS_1, VTU_OPS_REGISTER, &x) < 0) {
				fprintf(stderr, "Timeout %s %d\n", __FILE__, __LINE__);
				return -1;
			}

			if (! (x & (1 << 15)))  /* wait for VB=0 in Global Reg 0x05 */
			break;
		}

		if (phy_read(GLOBAL_REGS_1, VTU_VID_REG, &x) < 0) {
			fprintf(stderr, "Timeout %s %d\n", __FILE__, __LINE__);
			return -1;
		}

		if ((x & 0xFFF) == 0xFFF)
		break;

		j++;

		if (phy_read(GLOBAL_REGS_1, 0x07, &r7) < 0) {
			fprintf(stderr, "Timeout %s %d\n", __FILE__, __LINE__);
			return -1;
		}

		if (phy_read(GLOBAL_REGS_1, 0x08, &r8) < 0) {
			fprintf(stderr, "Timeout %s %d\n", __FILE__, __LINE__);
			return -1;
		}

		for(p=0; p < 7; p++) {
			switch(p) {
			  case 0: if ((r7 & 0x3) != 0x3)  goto L1; break;
			  case 1: if ((r7 & 0x30) != 0x30)  goto L1; break;
			  case 2: if ((r7 & 0x300) != 0x300)  goto L1; break;
			  case 3: if ((r7 & 0x3000) != 0x3000)  goto L1; break;
			  case 4: if ((r8 & 0x3) != 0x3)  goto L1; break;
			  case 5: if ((r8 & 0x30) != 0x30)  goto L1; break;
			  case 6: if ((r8 & 0x300) != 0x300)  goto L1; break;
			}

			continue;

L1:
			for(i=0; i < 7; i++) {
				if (i != p) {  /* don't set "our" bit in "our" mask register */
					switch(i) {
					  case 0: if ((r7 & 0x3) != 0x3) { r6[p] |= (1 << i);  } break;
					  case 1: if ((r7 & 0x30) != 0x30) { r6[p] |= (1 << i); } break;
					  case 2: if ((r7 & 0x300) != 0x300) { r6[p] |= (1 << i); } break;
					  case 3: if ((r7 & 0x3000) != 0x3000) { r6[p] |= (1 << i);  }  break;
					  case 4: if ((r8 & 0x3) != 0x3) { r6[p] |= (1 << i);  } break;
					  case 5: if ((r8 & 0x30) != 0x30) { r6[p] |= (1 << i); } break;
					  case 6: if ((r8 & 0x300) != 0x300) { r6[p] |= (1 << i); } break;
					}
				}
				else if (p != cpu_port) {
					if (entry->tags[p] != 3) {
						if (entry->forceVID[p])
						phy_write(0x18 + p, 0x7, 0x1000 | entry->vid);
						else
						phy_write(0x18 + p, 0x7, entry->vid);
					}
				}
			}
		}
	}

	for(p=0; p < 7; p++) {
		phy_write(0x18 + p, 0x6, r6[p]);

		if (entry->tags[p] != 3)
		phy_write(0x18 + p, 0x8, 0x480);
	}

	return 0;
}

static void usage(char **argv) {
	fprintf(stderr, "Usage: %s [OPTION] ...\n"
	  "Technologic Systems TS-76xx/TS-46xx/TS-74xx manipulation.\n"
	  "\n"
	  "General options:\n"
	  "  -g, --getmac            Display ethernet MAC address\n"
	  "  -s, --setmac=MAC        Set ethernet MAC address\n"
	  "  -R, --reboot            Reboot the board\n"
	  "  -t, --getrtc            Set system time from RTC\n"
	  "  -S, --setrtc            Set RTC time from system\n"
	  "  -o, --rtcinfo           Print RTC info, batt, temp, etc.\n"
	  "  -v, --nvram             Get/Set RTC NVRAM\n"
	  "  -i, --info              Display board info\n"
	  "  -e, --greenledon        Turn green LED on\n"
	  "  -b, --greenledoff       Turn green LED off\n"
	  "  -c, --redledon          Turn red LED on\n"
	  "  -d, --redledoff         Turn red LED off\n"
	  "  -X, --resetswitchon     Enable reset switch\n"
	  "  -Y, --resetswitchoff    Disable reset switch\n"
	  "  -D, --setdio=<pin>      Sets DDR and sets specified pin(s)\n"
	  "  -O, --clrdio=<pin>      Sets DDR and clears specified pin(s)\n"
	  "  -G, --getdio=<pin>      Clears DDR and gets DIO pin(s) input value\n"
	  "  -W, --watchdog          Daemonize and set up /dev/watchdog\n"
	  "  -q, --cputemp           Print CPU internal and external temperature\n"
	  "  -h, --help              This help\n",
	  argv[0]
	);
	
	if(model == 0x7600) {
		fprintf(stderr,
		  "  -N, --canon             Enable CAN on DIO pins\n"
		  "  -f, --canoff            Set CAN pins to DIO mode\n"
		);  
	}

	if(!is_7400()) {
		fprintf(stderr,
		  "  -a, --address=<adr>     8bit address to use with peek16/poke16\n"
		  "  -r, --peek16            Read 16bit register at <adr>\n"
		  "  -w, --poke16=<dat>      Write 16bit <dat> to <adr>\n"
		  "  -U, --removejp=JP       Remove soft jumper numbered JP (1-8)\n"
		  "  -J, --setjp=JP          Set soft jumper numbered JP (1-8)\n"
		  "  -k, --txenon=XUART(s)   Enables the TX Enable for an XUART\n"
		  "  -K, --txenoff=XUART(s)  Disables a specified TX Enable\n"
		  "  -x, --random            Get 16-bit hardware random number\n"
		  "  -n, --setrng            Seed the kernel random number generator\n"
	 	  "  -l, --loadfpga=FILE     Load FPGA bitstream from FILE\n"
		);
	} else {
		fprintf(stderr,
		  "  -M, --timewkup=<time>   Time in seconds to wake up after\n"
		  "  -m, --resetswitchwkup   Wake up at reset switch is press\n"
		  "  -F, --standby           Standby CPU, must specify -M|-m to wkup\n"
		  "  -L, --sleep             Sleep CPU, must specify -M|-m to wkup\n"
		  "  -N, --canon             Enable CAN transcievers\n"
		  "  -f, --canoff            Disable CAN transcievers\n"
		  "  -V, --cpuadc            Read and print CPU LRADC values\n"
		);
		if(model == 0x7670) {
		fprintf(stderr, 
		  "  -4, --485speed=<baud>   Baud rate, required for RS-485HD to function\n"
		  "  -1, --modbuspoweron     Enable VIN to modbus port\n"
		  "  -Z, --mosbuspoweroff    Turn off VIN to modbus port\n"
		  "  -2, --232on             Enable RS-232 transciever\n"
		  "  -3, --232off            Disable RS-232 transciever to save power\n"
		);
		}
		if(model == 0x7680) {
		fprintf(stderr,
		  "  -8, --485_0speed=<baud> Baud rate for RS-485 port 0 (UART 2)\n"
		  "  -9, --485_1speed=<baud> Baud rate for RS-485 port 1 (UART 3)\n"
		  "  -1, --modbuspoweron     Enable VIN to modbus port\n"
		  "  -Z, --mosbuspoweroff    Turn off VIN to modbus port\n"
		  "  -a, --address=<adr>     8bit address to use with peek16/poke16\n"
		  "  -r, --peek16            Read 16bit data at <adr> in EEPROM\n"
		  "  -w, --poke16=<dat>      Write 16bit <dat> to <adr> in EEPROM\n"
		  "  -6, --peekf             Read 8bit data at <adr> in FPGA\n"
		  "  -7, --pokef=<dat>       Write 8bit <dat> to <adr> in FPGA\n"
		  "      --dac0=<PWMval>     Write <PWMval> to DAC 0\n"
		  "      --dac1=<PWMval>     Write <PWMval> to DAC 1\n"
		  "      --dac2=<PWMval>     Write <PWMval> to DAC 2\n"
		  "      --dac3=<PWMval>     Write <PWMval> to DAC 3\n"
		);
		}
	}

	if(model == 0x4600) {
		fprintf(stderr,
		  "  -j, --bbclkon           Enables a 12.5MHz clock on DIO 3\n"
		  "  -E, --bbclk2on          Enables a 25MHz clock on DIO 37\n"
		  "  -I, --bbclk3on          Enables a 14.28MHz clock on DIO 3\n"
		  "  -H, --bbclkoff          Disables all baseboard clocks\n"
		  "  -u, --touchon           Turns the touchscreen controller on\n"
		  "  -T, --touchoff          Turns the touchscreen controller off\n"
		  "  -B, --baseboard         Display baseboard ID\n"
		  "  -A, --adc               Display MCP3428 ADC readings in millivolts\n"
		  "  -P, --ethvlan           Set each switch port to its own VLAN\n"
		  "  -y, --ethswitch         Set all switch ports to switch mode\n"
		  "  -5, --ethwlan           Sets port A on VLAN, other ports to switch mode\n"
		  "  -C, --ethinfo           Retrieves info on the onboard switch\n"
		  "  -Q  --ethbb             --eth* flags above are issued to baseboard switch\n"
		  );
	}
}

static inline
unsigned char tobcd(unsigned int n) {
	unsigned char ret;
	ret = (n / 10) << 4;
	ret |= n % 10;
	return ret;
}

static inline unsigned int frombcd(unsigned char x) {
	unsigned int ret;
	ret = (x >> 4) * 10;
	ret += x & 0xf;
	return ret;
}

int main(int argc, char **argv) {
	int devmem, i, c;
	unsigned int tagmem[24];
	unsigned int baseboard = 0;
	volatile unsigned short swmod;
	int opt_peek16 = 0, opt_address = 0, opt_poke16 = 0, opt_pokeval = 0;
	int opt_getmac = 0, opt_setmac = 0, opt_reboot = 0, opt_getrtc = 0;
	int opt_setrtc = 0, opt_info = 0, opt_greenledon = 0;
	int opt_greenledoff = 0, opt_redledon = 0, opt_redledoff = 0;
	int opt_setdio = 0, opt_clrdio = 0, opt_getdio = 0;
	int opt_watchdog = 0, opt_resetswitchon = 0, opt_resetswitchoff = 0;
	int opt_cputemp = 0, opt_random = 0, opt_setrng = 0;
	int opt_removejp = 0, opt_setjp = 0, opt_232 = 0;
	int opt_rtcinfo = 0, opt_nvram = 0;
	int opt_touchon = 0, opt_touchoff = 0;
	int opt_bbclkon = 0, opt_bbclkoff = 0, opt_bbclk2on=0, opt_bbclk3on=0;
	int opt_txenoff = 0, opt_txenon = 0, opt_can = 0;
	int opt_ethvlan = 0, opt_ethswitch = 0, opt_ethinfo = 0, opt_ethwlan=0;
	int opt_sleepmode = 0, opt_resetswitchwkup = 0;
	int opt_extwkup = 0, opt_timewkup = 0, opt_485speed = 0;
	int opt_baseboard = 0, opt_adc = 0, opt_modbuspoweron = 0;
	int opt_modbuspoweroff = 0, opt_cpuadc = 0;
	int opt_pokef = 0, opt_peekf = 0, opt_485_0speed = 0, opt_485_1speed = 0;
	int opt_dac0 = 0, opt_dac1 = 0, opt_dac2 = 0, opt_dac3 = 0;
	char *setdio = NULL, *clrdio = NULL, *getdio = NULL, *txen_on = NULL;
	char *txen_off = NULL;
	char *opt_mac = NULL, *opt_loadfpga = NULL;
	static struct option long_options[] = {
	  { "modbuspoweron", 0, 0, '1'},
	  { "modbuspoweroff", 0, 0, 'Z'},
	  { "485speed", 1, 0, '4'},
	  { "232en", 0, 0, '2'},
	  { "232dis", 0, 0, '3'},
	  { "232on", 0, 0, '2'},
	  { "232off", 0, 0, '3'},
	  { "sleep", 0, 0, 'L' },
	  { "standby", 0, 0, 'F' },
	  { "resetswitchwkup", 0, 0, 'm' },
	  { "cpuadc", 0, 0, 'V' },
	  { "timewkup", 1, 0, 'M' },
	  { "help", 0, 0, 'h' },
	  { "peek16", 0, 0, 'r' },
	  { "poke16", 1, 0, 'w' },
	  { "address", 1, 0, 'a' },
	  { "getmac", 0, 0, 'g' },
	  { "setmac", 1, 0, 's' },
	  { "reboot", 0, 0, 'R' },
	  { "getrtc", 0, 0, 't' },
	  { "setrtc", 0, 0, 'S' },
	  { "rtcinfo", 0, 0, 'o' },
	  { "greenledon", 0, 0, 'e'},
	  { "greenledoff", 0, 0, 'b'},
	  { "redledon", 0, 0, 'c'},
	  { "redledoff", 0, 0, 'd'},
	  { "setdio", 1, 0, 'D'},
	  { "clrdio", 1, 0, 'O'},
	  { "getdio", 1, 0, 'G'},
	  { "info", 0, 0, 'i' },
	  { "baseboard", 0, 0, 'B'},
	  { "random", 0, 0, 'x' },
	  { "watchdog", 0, 0, 'W' },
	  { "setrng", 0, 0, 'n' },
	  { "resetswitchon", 0, 0, 'X' },
	  { "resetswitchoff", 0, 0, 'Y' },
	  { "loadfpga", 1, 0, 'l' },
	  { "cputemp", 0, 0, 'q' },
	  { "removejp", 1, 0, 'U' },
	  { "setjp", 1, 0, 'J' },
	  { "nvram", 0, 0, 'v' },
	  { "touchon", 0, 0, 'u' },
	  { "touchoff", 0, 0, 'T' },
	  { "bbclkon", 0, 0, 'j' },
	  { "bbclk2on", 0, 0, 'E' },
	  { "bbclk3on", 0, 0, 'I' },
	  { "bbclkoff", 0, 0, 'H' },
	  { "txenon", 1, 0, 'k' },
	  { "txenoff", 1, 0, 'K' },
	  { "canon", 0, 0, 'N' },
	  { "canoff", 0, 0, 'f' },
	  { "ethvlan", 0, 0, 'P' },
	  { "ethswitch", 0, 0, 'y' },
	  { "ethinfo", 0, 0, 'C' },
	  { "ethwlan", 0, 0, '5' },
	  { "adc", 0, 0, 'A' },
	  { "ethbb", 0, 0, 'Q' },
	  { "peekf", 0, 0, '6' },
	  { "pokef", 1, 0, '7' },
	  { "485_0speed", 1, 0, '8' },
	  { "485_1speed", 1, 0, '9' },
	  { "dac0", 1, 0, 0 },
	  { "dac1", 1, 0, 1 },
	  { "dac2", 1, 0, 2 },
	  { "dac3", 1, 0, 3 },
	  { 0, 0, 0, 0 }
	};

	devmem = open("/dev/mem", O_RDWR|O_SYNC);
	assert(devmem != -1);
	mxgpioregs = mmap(0, getpagesize(), PROT_READ|PROT_WRITE, MAP_SHARED,
	  devmem, 0x80018000);

	if(!is_7400()) {
		
		nbuslock();
		mxgpioregs[0x104/4] = 0x0000ffff; //Set pinmux to GPIO
		mxgpioregs[0x114/4] = 0x003f0c03; //Set pinmux to GPIO
		mxgpioregs[0x704/4] = 0x070100ff; //Set to logic high
		mxgpioregs[0xb00/4] = 0x070100ff; //Set to output
	
		model = peek16(0);
		nbusunlock();
	}

	while((c = getopt_long(argc, argv,
	  "1234:567:8:9:ABCD:EFG:HIJ:K:LM:N:O:PQ:RSTU:VWXYZa:bcdef:ghijk:l:mnoqrs:tuvw:xy",
	  long_options, NULL)) != -1) {
		switch (c) {
		case 0:
			opt_dac0 = ((strtoul(optarg, NULL, 0) & 0xfff) << 1) | 0x1;
			break;
		case 1: 
			opt_dac1 = ((strtoul(optarg, NULL, 0) & 0xfff) << 1) | 0x1;
			break;
		case 2:
			opt_dac2 = ((strtoul(optarg, NULL, 0) & 0xfff) << 1) | 0x1;
			break;
		case 3: 
			opt_dac3 = ((strtoul(optarg, NULL, 0) & 0xfff) << 1) | 0x1;
			break;
		case '2':
			opt_232 = 1;
			break;
		case '3':
			opt_232 = 2;
			break;
		case '4':
			opt_485speed = strtoul(optarg, NULL, 0);
			break;
		case '6':
			opt_peekf = 1;
			break;
		case '7':
			opt_pokef = 1;
			opt_pokeval = strtoul(optarg, NULL, 0);
			break;
		case '8':
			opt_485_0speed = strtoul(optarg, NULL, 0);
			break;
		case '9':
			opt_485_1speed = strtoul(optarg, NULL, 0);
			break;
		case '1':
			opt_modbuspoweron = 1;
			break;
		case 'Z':
			opt_modbuspoweroff = 1;
			break;
		case 'M':
			opt_timewkup = strtoul(optarg, NULL, 0);
			break;
		case 'm':
			opt_resetswitchwkup = 1;
			break;
		case 'V':
			opt_cpuadc = 1;
			break;
		case 'F':
			opt_sleepmode = 2;
			break;
		case 'L':
			opt_sleepmode = 1;
			break;
		case 'v':
			opt_nvram = 1;
			break;
		case 'q':
			opt_cputemp = 1;
			break;
		case 'n':
			opt_setrng = 1;
			break;
		case 'l':
			opt_loadfpga = strdup(optarg);
			break;
		case 'X':
			opt_resetswitchon = 1;
			break;
		case 'Y':
			opt_resetswitchoff = 1;
			break;
		case 'W':
			opt_watchdog = 1;
			break;
		case 'D':
			opt_setdio = 1;	
			setdio = strdup(optarg);
			break;
		case 'O':
			opt_clrdio = 1;
			clrdio = strdup(optarg);
			break;
		case 'G':
			opt_getdio = 1;
			getdio = strdup(optarg);
			break;
		case 'e':
			opt_greenledon = 1;
			break;
		case 'b':
			opt_greenledoff = 1;
			break;
		case 'c':
			opt_redledon = 1;
			break;
		case 'd':
			opt_redledoff = 1;
			break;
		case 'i':
			opt_info = 1;
			break;
		case 'B':
			opt_baseboard = 1;
			break;
		case 'S':
			opt_setrtc = 1;
			break;
		case 'r':
			opt_peek16 = 1;
			break;
		case 'a':
			opt_address = strtoul(optarg, NULL, 0);
			break;
		case 'w':
			opt_poke16 = 1;
			opt_pokeval = strtoul(optarg, NULL, 0);
			break;
		case 'g':
			opt_getmac = 1;
			break;
		case 's':
			opt_setmac = 1;
			opt_mac = strdup(optarg);
			break;
		case 'R':
			opt_reboot = 1;
			break;
		case 't':
			opt_getrtc = 1;
			break;
		case 'o':
			opt_rtcinfo = 1;
			break;
		case 'x':
			opt_random = 1;
			break;
		case 'U':
			opt_removejp = strtoull(optarg, NULL, 0);
			break;
		case 'J':
			opt_setjp = strtoull(optarg, NULL, 0);
			break;
		case 'T':
			opt_touchoff = 1;
			break;
		case 'u':
			opt_touchoff = 1;
			opt_touchon = 1;
			opt_baseboard = 1;
			break;
		case 'j':
			opt_bbclkon = 1;
			opt_bbclkoff = 1;
			break;
		case 'E':
			opt_bbclk2on = 1;
			opt_bbclkoff = 1;
			break;
		case 'I':
			opt_bbclk3on = 1;
			opt_bbclkoff = 1;
			break;
		case 'H':
			opt_bbclkoff = 1;
			break;
		case 'k':
			opt_txenon = 1;
			txen_on = strdup(optarg);
			break;
		case 'K':
			opt_txenoff = 1;
			txen_off = strdup(optarg);
			break;
		case 'N':
			opt_can = 1;
			break;
		case 'f':
			opt_can = 2;
			break;
		case 'P':
			opt_ethvlan = 1;
			break;
		case 'y':
			opt_ethswitch = 1;
			break;
		case 'C':
			opt_ethinfo = 1;
			break;
		case '5':
			opt_ethwlan = 1;
			break;
		case 'A':
			opt_baseboard = 1;
			opt_adc = 1;
			break;
		case 'Q':
			opt_ethbb = 1;
			break;
		case 'h':
		default:
			usage(argv);
			return(1);
		}
	}
	
	if (opt_loadfpga) {
		unsigned int prev_gpio_en, prev_gpio_ddr, prev_gpio_mux;
		signed char x;
		const char * ispvmerr[] = { "pass", "verification fail",
		  "can't find the file", "wrong file type", "file error",
		  "option error", "crc verification error" };
		
		if(model == 0x7680) {
			/* Switch SPI pins to SPI mode */
			prev_gpio_mux = mxgpioregs[0x150/4];
			prev_gpio_en = mxgpioregs[0x720/4];
			prev_gpio_ddr = mxgpioregs[0xb20/4];

			/* Put most SPI pins to SPI mode */
			//mosi, miso, ss0
			mxgpioregs[0x158/4] = 0xff;
			//clk stay as GPIO for now
			mxgpioregs[0x154/4] = 0x3;
			mxgpioregs[0xb24/4] = 0x10000; //set clk output
			/* Leave SPI_RESET and CS in GPIO mode */
			//mxgpioregs[0x154/4] = 0xf00;
			//Set reset, done, and CS to gpio, even though they already should be
			mxgpioregs[0x134/4] = 0x3f;
			/* Set FPGA_DONE pin as input, reset and CS as output */
			mxgpioregs[0xb18/4] = 0x10000;
			mxgpioregs[0xb14/4] = 0x60000;

			/* The iCE40 chip needs to have the following done to
			 * put it in the mode for us to program it:
			 *
			 * CS pulled low
			 * FPGA_RESET pulled low
			 * SPI_CLK pulled high
			 * Wait at least 200ns
			 * FPGA_RESET pulled high
			 * Wait at least 200us

			 * Start sending data, MSB, falling edge, 1MHz to 25MHz
			 * Send minimum of 49 additional clocks
			 * Check FPGA_DONE, 1 == done
			 *
			 * Manual notes in bold and multiple times, that the
			 * SPI data stream must be uninterrupted, no idea what
			 * that actually means...
			 */

			x = ispSPI(opt_loadfpga);
			nbusunlock();
			if(x == 0) {
				printf("loadfpga_ok=1\n");
			} else {
				printf("loadfpga_ok=0\n");
				printf("loadfpga_error=\"%s\"\n", ispvmerr[-x]);
			}

			mxgpioregs[0xb20/4] = prev_gpio_ddr;
			mxgpioregs[0x720/4] = prev_gpio_en;
			mxgpioregs[0x150/4] = prev_gpio_mux;

		} else if(!is_7400()) {

			mxtimerregs = (unsigned int *) mmap(0, getpagesize(),
			  PROT_READ | PROT_WRITE, MAP_SHARED, devmem, 0x80068000);
			prev_gpio_mux = mxgpioregs[0x160/4];
			prev_gpio_en = mxgpioregs[0x730/4];
			prev_gpio_ddr = mxgpioregs[0xb30/4];
			/* switch JTAG pins into GPIO mode */
			mxgpioregs[0x160/4] |= 0xF;
			/* Set dout, clk, tms as outputs */
			mxgpioregs[0xb30/4] |= 0xD;
			mxgpioregs[0xb30/4] &= ~(0x2); /* Din as inputs */
	
			x = ispVM(opt_loadfpga);
			nbusunlock();
			if (x == 0) {
				printf("loadfpga_ok=1\n");
			} else {
				assert(x < 0);
				printf("loadfpga_ok=0\n");
				printf("loadfpga_error=\"%s\"\n", ispvmerr[-x]);
			}
	
			mxgpioregs[0xb30/4] = prev_gpio_ddr;
			mxgpioregs[0x730/4] = prev_gpio_en;
			mxgpioregs[0x160/4] = prev_gpio_mux;
		}
	}
	nbuslock();

	if (opt_cputemp) {
		signed int temp[2] = {0,0}, x;
		volatile unsigned int *mxlradcregs;

		nbusunlock();
		mxlradcregs = (unsigned int *) mmap(0, getpagesize(),
		  PROT_READ | PROT_WRITE, MAP_SHARED, devmem, 0x80050000);

		if(!is_7400()) {
			mxlradcregs[0x148/4] = 0xf;
			mxlradcregs[0x144/4] = 0x0; //Set LRDAC0 to channel 0
			mxlradcregs[0x50/4] = 0x0; //Clear ch0 reg
			for(x = 0; x < 20; x++) {
				/*
				 * Clear interrupt
				 * Set to 20uA or 300uA mode and enable current source
				 * Start conversion
				 * Poll for sample completion
				 * Pull out samples
				 */
				mxlradcregs[0x18/4] = 0x1;
				if(x < 10) mxlradcregs[0x20/4] = 0x8101;
				else mxlradcregs[0x20/4] = 0x810F;
				mxlradcregs[0x04/4] = 0x1;
				while(!(mxlradcregs[0x10/4] & 0x1));
				if(x < 10) temp[0] += (mxlradcregs[0x50/4] & 0xffff);
				else temp[1] += (mxlradcregs[0x50/4] & 0xffff);
			}	
			mxlradcregs[0x28/4] = 0x8100;
			temp[0] = (((temp[1] - temp[0])*1104)-2730000) + 30000;
			//+30000 is for compensation of wire impedence
			printf("external_temp=%d.%d\n", temp[0] / 10000,
			  abs(temp[0] % 10000));
		}
		
		mxlradcregs[0x148/4] = 0xFF;
		mxlradcregs[0x144/4] = 0x98; //Set to temp sense mode
		mxlradcregs[0x28/4] = 0x8300; //Enable temp sense block
		mxlradcregs[0x50/4] = 0x0; //Clear ch0 reg
		mxlradcregs[0x60/4] = 0x0; //Clear ch1 reg
		temp[0] = temp[1] = 0;

		for(x = 0; x < 10; x++) {
			/* Clear interrupts
			 * Schedule readings
			 * Poll for sample completion
			 * Pull out samples*/
			mxlradcregs[0x18/4] = 0x3;
			mxlradcregs[0x4/4] = 0x3;
			while(!((mxlradcregs[0x10/4] & 0x3) == 0x3)) ;
			temp[0] += mxlradcregs[0x60/4] & 0xFFFF;
			temp[1] += mxlradcregs[0x50/4] & 0xFFFF;
		}
		temp[0] = (((temp[0] - temp[1]) * (1012/4)) - 2730000);
		printf("internal_temp=%d.%d\n",temp[0] / 10000,
		  abs(temp[0] % 10000));
		nbuslock();
	}

	if(opt_poke16 && !is_7400()) {
		poke16(opt_address, opt_pokeval);
	}

	if(opt_peek16 && !is_7400()) {
		i = peek16(opt_address);
		nbusunlock();
		printf("0x%x\n", i);
		nbuslock();
	}

	if (opt_setmac) {
		/* This uses one time programmable memory. */
		unsigned int a, b, c;
		int r;
		volatile unsigned int *mxocotpregs;
		nbusunlock();

		r = sscanf(opt_mac, "%*x:%*x:%*x:%x:%x:%x",  &a,&b,&c);
		assert(r == 3); /* XXX: user arg problem */

		mxocotpregs = (unsigned int *) mmap(0, getpagesize(),
		  PROT_READ | PROT_WRITE, MAP_SHARED, devmem, 0x8002C000);

		mxocotpregs[0x08/4] = 0x200;
		mxocotpregs[0x0/4] = 0x1000;
		while(mxocotpregs[0x0/4] & 0x100) ; //check busy flag
		if(mxocotpregs[0x20/4] & (0xFFFFFF)) {
			printf("MAC address previously set, cannot set\n");
		} else {
			assert(a < 0x100);
			assert(b < 0x100);
			assert(c < 0x100);
			mxocotpregs[0x0/4] = 0x3E770000;
			mxocotpregs[0x10/4] = (a<<16|b<<8|c);
		}
		mxocotpregs[0x0/4] = 0x0;
		nbuslock();
	}

	if (opt_getmac) {
		unsigned char a, b, c;
		unsigned int mac, i;
		volatile unsigned int *mxocotpregs;
		mxocotpregs = (unsigned int *) mmap(0, getpagesize(),
		  PROT_READ | PROT_WRITE, MAP_SHARED, devmem, 0x8002C000);

		mxocotpregs[0x08/4] = 0x200;
		mxocotpregs[0x0/4] = 0x1000;
		while(mxocotpregs[0x0/4] & 0x100) ; //check busy flag
		mac = mxocotpregs[0x20/4] & 0xFFFFFF;
		if(!mac) {
			mxocotpregs[0x0/4] = 0x0; //Close the reg first
			mxocotpregs[0x08/4] = 0x200;
			mxocotpregs[0x0/4] = 0x1013;
			while(mxocotpregs[0x0/4] & 0x100) ; //check busy flag
			mac = (unsigned short) mxocotpregs[0x150/4];
			mac |= 0x4f0000;
		}
		mxocotpregs[0x0/4] = 0x0;

		a = mac >> 16;
		b = mac >> 8;
		c = mac;
		
		if(!is_7400()) {
			read_tagmem(tagmem);
		}
		nbusunlock();
		printf("mac=00:d0:69:%02x:%02x:%02x\n", a, b, c);
		printf("shortmac=%02x%02x%02x\n", a, b, c);
		printf("model=0x%x\n", model);

		if(!is_7400()) {
			for (i = 0; i < 8; i++) printf("jp%d=%c\n", i + 1,
			  ((tagmem[0] & (1<<i)) ? '0' : '1'));
		} else {
			for (i = 0; i < 8; i++) printf("jp%d=%c\n", i + 1,
			  '0');
		}
		nbuslock();
	}

	if ((opt_setjp || opt_removejp) && !is_7400()) {
		unsigned int origtag[24];
	        read_tagmem(origtag);
		memcpy(tagmem, origtag, 24 * sizeof (int));

	        if (opt_setjp) tagmem[0] &= (char)~(1<<(opt_setjp-1));
	        else tagmem[0] |= (char)(1<<(opt_removejp-1));
		if(memcmp(origtag, tagmem, 24 * sizeof (int))) {
	        	erase_tagmem();
		        write_tagmem(tagmem);
		}
	}
	
	if(opt_rtcinfo && model != 0x7680) {
		unsigned char rtcdat[11];

		rtc_read(0x28, &rtcdat, 2);
		nbusunlock();
		printf("rtctemp_millicelsius=%d\n",
		  ((rtcdat[0]|(rtcdat[1]<<8))*500)-273000);
		nbuslock();
		rtc_read(0x7, &rtcdat, 1);
		rtc_read(0x16, &rtcdat[1], 5);
		rtc_read(0x1b, &rtcdat[6], 5);

		nbusunlock();
		printf("rtcinfo_oscillator_ok=%d\n",
		  ((rtcdat[0] & 0x41) ? 0 : 1));
		printf("rtcinfo_batt_low=%d\n",
		  ((rtcdat[0] & 4) >> 2));
		printf("rtcinfo_batt_crit=%d\n",
		  ((rtcdat[0] & 2) >> 1));
		printf("rtcinfo_firstpoweroff=%02x%02x%02x%02x%02x\n",
		  rtcdat[5], rtcdat[4], (rtcdat[3] & 0x7f), rtcdat[2],
		  rtcdat[1]);
		printf("rtcinfo_lastpoweron=%02x%02x%02x%02x%02x\n",
		  rtcdat[10], rtcdat[9], (rtcdat[8] & 0x7f), rtcdat[7],
		  rtcdat[6]);
		nbuslock();
		rtcdat[0] = 0x80;
		rtc_write(0x9, &rtcdat, 1);
	}

			
	if (opt_setrtc && model != 0x7680) {
		unsigned char rtcdat[7];
		time_t now = time(NULL);
		struct tm *tm;

		rtc_read(0xd, &rtcdat, 1);
		rtcdat[0] = (0xc0 | (rtcdat[0] & 0x1f));
		rtcdat[1] = 0x51;
		rtcdat[2] = 0x0;
		rtcdat[3] = 0x24;
		
		rtc_write(0x8, &rtcdat[1], 3);
		rtc_write(0xd, &rtcdat[0], 1);
		
		nbusunlock();

		tm = gmtime(&now);
		rtcdat[0] = tobcd((tm->tm_sec)+1);
		rtcdat[1] = tobcd(tm->tm_min);
		rtcdat[2] = tobcd(tm->tm_hour) | 0x80; //24 hour bit
		rtcdat[3] = tobcd(tm->tm_mday);
		rtcdat[4] = tobcd(tm->tm_mon + 1);
		assert(tm->tm_year >= 100);
		rtcdat[5] = tobcd(tm->tm_year % 100);
		rtcdat[6] = tm->tm_wday + 1;

		nbuslock();
		
		rtc_write(0x0, &rtcdat, 7);
	}

	if (opt_getrtc && model != 0x7680) {
	        unsigned char rtcdat[8];
	        time_t now;
	        struct tm tm;
	        struct timeval tv;
	
		rtc_read(0x0, &rtcdat, 8);

		nbusunlock();
		printf("rtc_oscillator_ok=%d\n", rtcdat[7] & 1 ? 0 : 1);
	        if (!(rtcdat[7] & 1)) {
	                tm.tm_sec = frombcd(rtcdat[0] & 0x7f);
	                tm.tm_min = frombcd(rtcdat[1] & 0x7f);
	                tm.tm_hour = frombcd(rtcdat[2] & 0x3f);
	                tm.tm_mday = frombcd(rtcdat[3] & 0x3f);
	                tm.tm_mon = frombcd(rtcdat[4] & 0x1f) - 1;
	                tm.tm_year = frombcd(rtcdat[5]) + 100;
	                setenv("TZ", "UTC", 1);
	                now = mktime(&tm);
	                tv.tv_sec = now;
	                tv.tv_usec = 0;
	                settimeofday(&tv, NULL);
	        }
		nbuslock();
	}
	if (opt_nvram && model != 0x7680) {
		unsigned int dat[32], x;
	        unsigned char *cdat = (unsigned char *)dat;
		char *e, var[8];
	
		nvram_read(0x0, cdat, 128);
	        for (i = 0; i < 32; i++) {
	                sprintf(var, "nvram%d", i);
	                e = getenv(var);
	                if (e) {
	                        x = strtoul(e, NULL, 0);
				nvram_write(i<<2, &x, 4);
	                }
	        }
		nbusunlock();
	        for (i = 0; i < 32; i++) printf("nvram%d=0x%x\n", i, dat[i]);
		nbuslock();
	}


	if(opt_resetswitchon && !is_7400()) {
		unsigned short x;
		x = peek16(0x2);
		poke16(0x2, (x | (1 << 11)));
	} else if(opt_resetswitchoff && !is_7400()) {
		unsigned short x;
		x = peek16(0x2);
		poke16(0x2, (x & ~(1 << 11)));
	}

	if(opt_resetswitchon && is_7400()) {
		unsigned char dat = 0x42;
		m0_write(&dat, 1);
	} else if(opt_resetswitchoff && is_7400()) {
		unsigned char dat = 0x40;
		m0_write(&dat, 1);
	}

	if(opt_info && !is_7400()) {
		unsigned short x, y;
		x = peek16(0x0);
		y = peek16(0x2);
		nbusunlock();
		printf("model=0x%x\n", x);
		printf("submodel=0x%x\n", (y >> 4) & 0xf);
		printf("revision=0x%x\n", (y & 0xf));
		printf("bootmode=0x%x\n", (y >> 8) & 0x1);
		printf("fpga_pn=LFXP25E\n");
		printf("resetsw_en=0x%x\n", (y >> 11) & 0x1);
		nbuslock();
	} else if(opt_info && is_7400()) {
		unsigned char dat[18];
		m0_read(&dat, 17);
		if(model == 0x7680) fpga_read(0x7f, &dat[17], 1);
		nbusunlock();
		printf("model=0x%x\n", model);
		printf("bootmode=0x");
		if(model == 0x7680) printf("%x\n", (mxgpioregs[0x910/4] >> 12) & 0x1);
		else printf("%x\n", (dat[15] >> 3) & 0x1);
		printf("revision=0x%x\n", (dat[16] >> 4) & 0xf);
		if(model == 0x7680) printf("fpga_revision=0x%x\n", dat[17]);
		printf("reboot_source=");
		switch(dat[15] & 0x7) {
		  case 0:
			printf("poweron\n");
			break;
		  case 1:
			printf("WDT\n");
			break;
		  case 2:
			printf("resetswitch\n");
			break;
		  case 3:
			printf("sleep\n");
			break;
		  case 4:
			printf("uart_break\n");
			break;
		}
		printf("VIN_millivolts=");
		if(model == 0x7680) printf("%d\n", ((dat[0]) << 8 | dat[1])*12/10);
		else printf("%d\n", ((dat[0]+2) << 8 | dat[1]));
		nbuslock();
	}

	if(opt_sleepmode) {
		unsigned char dat[4] = {0};

		dat[0]=(0x1 | (opt_resetswitchwkup << 1) | (opt_extwkup << 2) |
		  ((opt_sleepmode-1) << 4) | 1 << 6);
		dat[3] = (opt_timewkup & 0xff);
		dat[2] = ((opt_timewkup >> 8) & 0xff);
		dat[1] = ((opt_timewkup >> 16) & 0xff);
		m0_write(&dat, 4);
	}

	if (opt_reboot) {
		struct sched_param sched;
		unsigned char dat[3] = {0, 0, 0};
		sync();
		if(is_7400()) {
			m0_write(&dat, 3);
		} else poke16(0x2a, 0);
		sched.sched_priority = 99;
		sched_setscheduler(0, SCHED_FIFO, &sched);
		while(1);
	}

	if(opt_setdio && !is_7400()) {
		int j = 0, sz = strlen(setdio);
		char *st = 0, *en = 0;
		for(st = setdio; j < sz; st = en + 1) {
			int dio = strtoul(st, &en, 0);
			j += (en - st) + 1;
			poke16(0x26, (0x80 | dio));
			poke16(0x28, (0x80 | dio));
		}
	} else if(opt_setdio && is_7400()) {
		int pin, bank, reg, muxpin;
		char *en = 0;
		

		setdio = strtok(setdio, ",");
		while(setdio != NULL) {
			bank = strtoul(setdio, &en, 0);
			nbusunlock();
			assert(bank < 5 && bank >= 0);
			nbuslock();
			pin = strtol((en+1), 0, 0);
			if(pin > 15) {
				reg = 1;
				muxpin = pin - 16;
			} else {
				reg = 0;
				muxpin = pin;
			}
			/* Set pin as GPIO */
			mxgpioregs[((0x100) + (0x20 * bank) + (0x10 * reg) +
			  (0x4))/4] = (0x3 << (muxpin * 2)); 
			/* Set data */
			mxgpioregs[((0x700) + (0x10 * bank) + (0x4))/4] =
			  (0x1 << pin);
			/* Set pin to output */
			mxgpioregs[((0xB00) + (0x10 * bank) + (0x4))/4] =
			  (0x1 << pin);
			setdio = strtok(NULL, ",");
		}
	}

	if(opt_clrdio && !is_7400()) {
		int j = 0, sz = strlen(clrdio);
		char *st = 0, *en = 0;
		for(st = clrdio; j < sz; st = en + 1) {
			int dio = strtoul(st, &en, 0);
			j += (en - st) + 1;
			poke16(0x26, (~0x80 & dio));
			poke16(0x28, (0x80 | dio));
		}
	} else if(opt_clrdio && is_7400()) {
		int pin, bank, reg, muxpin;
		char *en = 0;

		clrdio = strtok(clrdio, ",");
		while(clrdio != NULL) {
			bank = strtoul(clrdio, &en, 0);
			nbusunlock();
			assert(bank < 5 && bank >= 0);
			nbuslock();
			pin = strtol((en+1), 0, 0);
			if(pin > 15) {
				reg = 1;
				muxpin = pin - 16;
			} else {
				reg = 0;
				muxpin = pin;
			}
			/* Set pin as GPIO */
			mxgpioregs[((0x100) + (0x20 * bank) + (0x10 * reg) +
			   (0x4))/4] = (0x3 << (muxpin * 2)); 
			/* Clr data */
			mxgpioregs[((0x700) + (0x10 * bank) + (0x8))/4] =
			  (0x1 << pin);
			/* Set pin to output */
			mxgpioregs[((0xB00) + (0x10 * bank) + (0x4))/4] =
			  (0x1 << pin);
			clrdio = strtok(NULL, ",");
		}
	}

	if(opt_getdio && !is_7400()) {
		int j = 0, sz = strlen(getdio);
		char *st = 0, *en = 0;
		for(st = getdio; j < sz; st = en + 1) {
			int dio = strtoul(st, &en, 0);
			int value = -1;
			j += (en - st) + 1;
			poke16(0x28, (~0x80 & dio));
			if(dio < 16)
			  value = (peek16(0x08) >> (dio));
			else if (dio < 32)
			  value = (peek16(0x0e) >> (dio - 16));
			else if (dio < 48)
			  value = (peek16(0x14) >> (dio - 32));
			else if (dio < 64)
			  value = (peek16(0x1a) >> (dio - 48));
			else if (dio < 70)
			  value = (peek16(0x20) >> (dio - 64));
			nbusunlock();
			printf("dio%d=%d\n", dio, value & 0x1);
			nbuslock();
		}
	} else if(opt_getdio && is_7400()) {
		int pin, bank, reg, muxpin;
		char *en = 0;

		getdio = strtok(getdio, ",");
		while(getdio != NULL) {
			bank = strtoul(getdio, &en, 0);
			nbusunlock();
			assert(bank < 5 && bank >= 0);
			pin = strtol((en+1), 0, 0);
			if(pin > 15) {
				reg = 1;
				muxpin = pin - 16;
			} else {
				reg = 0;
				muxpin = pin;
			}
			/* Set pin as GPIO */
			mxgpioregs[((0x100) + (0x20 * bank) + (0x10 * reg) +
			   (0x4))/4] = (0x3 << (muxpin * 2));
			/* Set pin to input */
			mxgpioregs[((0xB00) + (0x10 * bank) + (0x8))/4] =
			  (0x1 << pin);
			printf("dio%d_%d=%d\n", bank, pin,
			  (((mxgpioregs[((0x900) + (0x10 * bank))/4]) >> pin) &
			  0x1));
			nbuslock();
			getdio = strtok(NULL, ",");
		}
	}

	if(opt_232 && model == 0x7670) {
		mxgpioregs[(0x134)/4] = (0x3 << 9); 
		if(opt_232 == 1)
		  mxgpioregs[(0x714)/4] = (0x1 << 25);
		else
		  mxgpioregs[(0x718)/4] = (0x1 << 25);
		mxgpioregs[(0xB14)/4] = (0x1 << 25);
	}

	if(opt_can && is_7400()) {
		mxgpioregs[(0x174)/4] = (0x3 << 28); 
		if(opt_can == 1)
		  mxgpioregs[(0x738)/4] = (0x1 << 30);
		else
		  mxgpioregs[(0x734)/4] = (0x1 << 30);
		mxgpioregs[(0xB34)/4] = (0x1 << 30);
	}

	if(opt_can && model == 0x7600) {
		if(opt_can == 1) 
		  poke16(0x2, (peek16(0x2) | (1 << 10)));
		else
		  poke16(0x2, (peek16(0x2) & ~(1 << 10)));
	}

	if(opt_random && !is_7400()) {
		unsigned short rand = peek16(0x4);
		nbusunlock();
		printf("random=0x%x\n", rand);
		nbuslock();
	}

	if (opt_setrng && !is_7400()) {
		FILE *urandom = NULL;
		unsigned int rng;
		rng = peek16(0x4);
		nbusunlock();
		urandom = fopen("/dev/urandom", "w");
		assert(urandom != NULL);
		fwrite(&rng, 2, 1, urandom);
		fclose(urandom);
		nbuslock();
	}

	if (opt_baseboard && model == 0x4600) {
		unsigned short prev1, prev2, prev3, prev4;
		unsigned int i, x, rev;
		int bb_fd = 0;
		char buf[5] = {0};

		bb_fd = open("/dev/tsbaseboardid", O_RDONLY);
		if(bb_fd > 0) {
			read(bb_fd, &buf, 4);
			buf[4] = '\0';
			baseboard = strtoul(buf, NULL, 0);
		} else {
			prev1 = peek16(0x2);
			prev2 = peek16(0x12);
			prev3 = peek16(0x10);
			prev4 = peek16(0xc);
			poke16(0x12, prev2 | 0x10);
			poke16(0xc, prev4 & ~0x20);
			for(i = 0; i < 8; i++) {
				x = prev1 & ~0xc000;
				if(!(i & 1)) x |= 0x4000;
				if(!(i & 2)) x |= 0x8000;
				if(i & 4) poke16(0x26, 20 | 0x80);
				else poke16(0x26, 20);
				poke16(0x2, x);
				usleep(1);
				baseboard = (baseboard >> 1);
				if(peek16(0x8) & 0x20) baseboard |= 0x80;
			}
			poke16(0x2, prev1);
			poke16(0x12, prev2);
			poke16(0x10, prev3);
			poke16(0xc, prev4);
		}

		rev = (baseboard & 0xc0) >> 6;
		baseboard &= ~0xc0;
		nbusunlock();
		printf("baseboard_model=0x%x\n", baseboard);
		switch (baseboard) {
		  /* The TS-8200 can ID as 63 on early production boards
		   * we do not want to ID by that as well as 63 could
		   * (more likely) be a custom baseboard
		   */
		  case 0:
			printf("baseboard=8200\n");
			break;
		  case 2:
			printf("baseboard=8390\n");
			break;
		  case 4:
			printf("baseboard=8500\n");
			break;
		  case 5:
			printf("baseboard=8400\n");
			break;
		  case 6:
			printf("baseboard=8160\n");
			break;
		  case 7:
			printf("baseboard=8100\n");
			break;
		  case 8:
			printf("baseboard=8820\n");
			break;
		  case 9:
			printf("baseboard=8150\n");
			break;
		  case 10:
			printf("baseboard=8900\n");
			break;
		  case 11:
			printf("baseboard=8290\n");
			break;
		  case 13:
			printf("baseboard=8700\n");
			break;
		  case 14:
			printf("baseboard=8280\n");
			break;
		  case 15:
			printf("baseboard=8380\n");
			break;
		  case 16:
			printf("baseboard=an20\n");
			break;
		}
		printf("baseboard_rev=%c\n", rev + 65);
		nbuslock();
	}

	if(opt_touchoff && model == 0x4600) poke16(0x2e, (peek16(0x2e) & ~0xc000));
	if(opt_touchon && model == 0x4600) {
		/* The TS-8380 is a little special and has different SPI pins */
		if(baseboard == 15) poke16(0x2e, (peek16(0x2e) | 0x8000));
		else poke16(0x2e, (peek16(0x2e) | 0x4000));
	}

	if (opt_cpuadc) {
		volatile unsigned int *mxlradcregs;
		volatile unsigned int *mxhsadcregs;
		volatile unsigned int *mxclkctrlregs;
		int x, chan[8] = {0,0,0,0,0,0,0,0};
		signed int bivolt;
		int i;

		mxlradcregs = (unsigned int *) mmap(0, getpagesize(),
		  PROT_READ | PROT_WRITE, MAP_SHARED, devmem, 0x80050000);

		mxlradcregs[0x148/4] = 0xfffffff; //Clear LRADC6:0 assignments
		mxlradcregs[0x144/4] = 0x6543210; //Set LRDAC6:0 to channel 6,5:0
		mxlradcregs[0x28/4] = 0xff000000; //Set 1.8v range
		for(x = 0; x < 7; x++)
		  mxlradcregs[(0x50+(x * 0x10))/4] = 0x0; //Clear LRADCx reg

		for(x = 0; x < 10; x++) {
			mxlradcregs[0x18/4] = 0x7f; //Clear interrupt ready
			mxlradcregs[0x4/4] = 0x7f; //Schedule conversaion of chan 6:0
			while(!((mxlradcregs[0x10/4] & 0x7f) == 0x7f)); //Wait
			for(i = 0; i < 7; i++)
			  chan[i] += (mxlradcregs[(0x50+(i * 0x10))/4] & 0xffff);
		}

		mxhsadcregs = mmap(0, getpagesize(), PROT_READ|PROT_WRITE, MAP_SHARED,
		  devmem, 0x80002000);
		mxclkctrlregs = mmap(0, getpagesize(), PROT_READ|PROT_WRITE, MAP_SHARED,
		  devmem, 0x80040000);

		//Lets see if we need to bring the HSADC out of reset
		if(mxhsadcregs[0x0/4] & 0xC0000000) {
			mxclkctrlregs[0x154/4] = 0x70000000;
			mxclkctrlregs[0x1c8/4] = 0x8000;
			//ENGR116296 errata workaround
			mxhsadcregs[0x8/4] = 0x80000000;
			mxhsadcregs[0x0/4] = ((mxhsadcregs[0x0/4] | 0x80000000) & (~0x40000000));
			mxhsadcregs[0x4/4] = 0x40000000;
			mxhsadcregs[0x8/4] = 0x40000000;
			mxhsadcregs[0x4/4] = 0x40000000;

			usleep(10);
			mxhsadcregs[0x8/4] = 0xc0000000;
		}

		mxhsadcregs[0x28/4] = 0x2000; //Clear powerdown
		mxhsadcregs[0x24/4] = 0x31; //Set precharge and SH bypass
		mxhsadcregs[0x30/4] = 0xa; //Set sample num
		mxhsadcregs[0x40/4] = 0x1; //Set seq num
		mxhsadcregs[0x4/4] = 0x40000; //12bit mode

		while(!(mxhsadcregs[0x10/4] & 0x20)) {
			mxhsadcregs[0x50/4]; //Empty FIFO
		}

		mxhsadcregs[0x50/4]; //An extra read is necessary

		mxhsadcregs[0x14/4] = 0xfc000000; //Clr interrupts
		mxhsadcregs[0x4/4] = 0x1; //Set HS_RUN
		usleep(10);
		mxhsadcregs[0x4/4] = 0x08000000; //Start conversion
		while(!(mxhsadcregs[0x10/4] & 0x1)) ; //Wait for interrupt

		for(i = 0; i < 5; i++) {
			x = mxhsadcregs[0x50/4];
			chan[7] += ((x & 0xfff) + ((x >> 16) & 0xfff));
		}


		nbusunlock();
		for(x = 0; x < 7; x++) {
			/* Multiply by 1000 to get mV, divide by 10 samples
			 * Analog dividers are divide by two on ADC 1-4,
			 * 6 is divide by 3. 45177 is the value of 
			 * (1.85/4095)x10^8 so we deal with ints
			 *
			 * Calibration offsets should be added here */		
			if(model == 0x7680) {
			  switch(x) {
				case 0:
				case 1:
				case 2:
				case 3:
				  chan[x] = ((((chan[x]/10)*45177)/1000000)*62);
				  printf("LRADC_ADC%d_millivolts=%d\n", x, chan[x]);
				  break;
				case 4:
				case 5:
				  /* This is done in two steps to clean up code
				   * The first step gets the mV vale read from
				   * ADC, and the second step applies Millman's
				   * theorem for the input circuit, simplified
				   * as much as possible.
				   */
				  bivolt=(((chan[x]/10)*45177)/100000);
				  bivolt=(((3*(((435937*bivolt)/1000)-412000))
				    /197));
				  printf("LRADC_ADC%d_millivolts=%d\n", x, bivolt);
				  break;
			  }
			} else {
			  switch(x) {
				case 1:
				case 2:
				case 3:
				case 4:
				  chan[x] = ((((chan[x]/10)*45177)/100000)*2);
				  printf("LRADC_ADC%d_millivolts=%d\n", x, chan[x]);
				  break;
				case 6:
				  chan[x] = ((((chan[x]/10)*45177)/1000000)*33);
				  printf("LRADC_ADC%d_millivolts=%d\n", x, chan[x]);
				  break;
			  }
			}			
				  
		}
		if(model != 0x7680) {
			printf("HSADC_millivolts=%d\n", 
			  ((((chan[7]/10)*45177)/100000)*2));
		}
		nbuslock();
	}

	if (opt_adc && model == 0x4600) {
		int x[6];
		// configure ADC:
		switch (baseboard) {
		  case 2:
			poke16(0x30, 0x28);
			break;
		  case 6:
		  case 7:
			poke16(0x30, 0x18);
			break;
		  default: // if unknown baseboard, uses no an_sel
			// but assumes ADC is there
			poke16(0x30, 0x08);
			break;
		}
		poke16(0x32, 0x3f); // enable all 6 channels
		nbusunlock();
		usleep(500000); // allow time for conversions
		nbuslock();
		for (i = 0; i <= 5; i++) {
			x[i] = (signed short)peek16(0x34 + 2*i);
		}
		nbusunlock();
		for(i = 0; i <= 5; i++) {
			if (i > 1) x[i] = (x[i] * 1006)/200;
			x[i] = (x[i] * 2048)/0x8000;
			printf("adc%d=%d\n", i+1, x[i]);
		}
		nbuslock();
	}

	if ((opt_ethswitch || opt_ethinfo || opt_ethvlan || opt_ethwlan) && has_switch()) {

		nbusunlock();
		mxgpioregs[0x184/4] = 0xf;

		if(phy_read(0x18, 0x03, &swmod) == -1) {
			printf("switch_model=none\n");
			opt_ethvlan = 0;
			opt_ethwlan = 0;
			opt_ethswitch = 0;
		} else {
			swmod &= 0xFFF0;
			if(swmod == 512){
				printf("switch_model=88E6020\n");
			}else if (swmod == 1792) {
				printf("switch_model=88E6070\n");
			} else if(swmod == 1808) {
				printf("switch_model=88E6071\n");
				fprintf(stderr, "Unsupported switch\n");
			} else if(swmod == 8704) {
				printf("switch_model=88E6220\n");
				fprintf(stderr, "Unsupported switch\n");
			} else if(swmod == 9472) {
				printf("switch_model=88E6251\n");
				fprintf(stderr, "Unsupported switch\n");
			} else {
				printf("switch_model=unknown\n");
			}
		}
		nbuslock();
	}

	if ((opt_ethswitch || opt_ethvlan || opt_ethwlan) && has_switch()) {
		nbusunlock();
		if(swmod == 512) { // Marvell 88E6020 Network Switch
			// this chip has 2 PHYs and 2 RMII ports
			// Apply Marvell 88E60x0 Errata 3.1 for PHY 0 and 1.
			switch_errata_3_1(0);
			switch_errata_3_1(1<<5);
			// enable both PHYs
			switch_enphy(0x9600);
			switch_enphy(0x9620);

			// enable port flood on all ports
			if(opt_ethswitch) {
				switch_enflood(0x18);
				switch_enflood(0x19);
				switch_enflood(0x1d);
			} else if (opt_ethvlan || opt_ethwlan) {
				switch_enbcastflood(0x18);
				switch_enbcastflood(0x19);
				switch_enbcastflood(0x1d);
			}

			// Force link up on P5, the CPU port
			switch_forcelink(0x1d);

			// With this chip ethvlan and wlan switch are the same since
			// we only have 2 ports
			if(opt_ethvlan || opt_ethwlan) {
				struct vtu entry;
				// Configure P5
				phy_write(0x1d, 0x05, 0x0);
				phy_write(0x1d, 0x06, 0x5f);
				phy_write(0x1d, 0x07, 0x1);
				phy_write(0x1d, 0x08, 0x480);

				for(i=0; i < 7; i++) {
					entry.tags[i] = 3; // leave the port alone by default
					entry.forceVID[i] = 0;
				}

				// Set up VLAN #1 on P0 and P5
				entry.v = 1;
				entry.vid = 1;
				entry.tags[5] = 2; // TS-4712 CPU port is always egress tagged
				entry.tags[0] = 1; // External Port A egress untagged
				vtu_add_entry(&entry);
				entry.tags[0] = 3; // Do not configure port A again

				// Setup VLAN #2 on P1 and P5
				entry.vid = 2;
				entry.tags[1] = 1; // External PORT B egress untagged
				vtu_add_entry(&entry);
			}
		} else if (swmod == 1792) { // Marvell 88E6070 Network Switch
			// Used on the TS-8700
			// enable all PHYs
			
			switch_errata_3_1(0);
			switch_errata_3_1(1<<5);
			switch_errata_3_1(2<<5);
			switch_errata_3_1(3<<5);
			switch_errata_3_1(4<<5);
			
			switch_enphy(0x9600);
			switch_enphy(0x9620);
			switch_enphy(0x9640);
			switch_enphy(0x9660);
			switch_enphy(0x9680);

			switch_enflood(0x18);
			switch_enflood(0x19);
			switch_enflood(0x1a);
			switch_enflood(0x1b);
			switch_enflood(0x1c);
			if(opt_ethvlan) {
				struct vtu entry;
				for(i=0; i < 7; i++) {
					entry.tags[i] = 3; // leave the port alone by default
					entry.forceVID[i] = 0;
				}
				entry.v = 1;

				// Set up VLAN #1 on P0 and P1
				entry.vid = 1;
				entry.tags[0] = 2; // TS-4710 CPU port is always egress tagged
				entry.tags[1] = 1; // External Port A egress untagged
				vtu_add_entry(&entry);
				entry.tags[1] = 3; // Do not configure port A again

				// Setup VLAN #2 on P0 and P2
				entry.vid = 2;
				entry.tags[2] = 1; // External PORT B egress untagged
				vtu_add_entry(&entry);
				entry.tags[2] = 3; // Do not configure port B again

				// Setup VLAN #2 on P0 and P3
				entry.vid = 3;
				entry.tags[3] = 1; // External PORT C egress untagged
				vtu_add_entry(&entry);
				entry.tags[3] = 3; // Do not configure port C again

				// Setup VLAN #2 on P0 and P4
				entry.vid = 4;
				entry.tags[4] = 1; // External PORT D egress untagged
				vtu_add_entry(&entry);
			} else if (opt_ethwlan) {
				struct vtu entry;
				for(i=0; i < 7; i++) {
					entry.tags[i] = 3; // leave the port alone by default
					entry.forceVID[i] = 0;
				}
				entry.v = 1;

				// Set up VLAN #1 with P0 and P1
				entry.vid = 1;
				entry.tags[0] = 2; // TS-4710 CPU port is always egress tagged
				entry.tags[1] = 1; // External Port A egress untagged
				vtu_add_entry(&entry);
				entry.tags[1] = 3; // Do not configure port A again

				// Set up VLAN #1 with P0, P2, P3, and P4
				entry.vid = 2;
				entry.tags[0] = 2; // TS-4710 CPU port is always egress tagged
				entry.tags[2] = 1; // External Port B egress untagged
				entry.tags[3] = 1; // External Port C egress untagged
				entry.tags[4] = 1; // External Port D egress untagged
				vtu_add_entry(&entry);
			}
		} else {
			fprintf(stderr, "Switch not configured\n");
		}
		nbuslock();
	}

	if (opt_ethinfo && has_switch()) {
		int ports = 0, i = 0, port, vtu = 0;
		int phy[8] = {0,0,0,0,0,0,0,0};
		volatile unsigned short x, r7;
		
		nbusunlock();

		if(swmod == 512){ // 4712
			ports = 2;
			phy[0] = 0x18;
			phy[1] = 0x19;
		} else if (swmod == 1792) { // 8700
			ports = 4;
			phy[0] = 0x19;
			phy[1] = 0x1a;
			phy[2] = 0x1b;
			phy[3] = 0x1c;
		} else {
			printf("Unknown switch: %d\n", swmod);
			return 1;
		}
		printf("switch_ports=\"");
		for (i = 0; i < ports; i++) {
			printf("%c", 97 + i);
			if (i != ports - 1)
			  printf(" ");
		}
		printf("\"\n");

		for (i = 0; i < ports; i++) {
			volatile unsigned short dat;
			phy_read(phy[i], 0x0, &dat);
			printf("switchport%c_link=%d\n", 97 + i,
			  dat & 0x1000 ? 1 : 0);
			printf("switchport%c_speed=", 97 + i);
			dat = (dat & 0xf00) >> 8;
			if(dat == 0x8)
			  printf("10HD\n");
			else if (dat == 0x9)
			  printf("100HD\n");
			else if (dat == 0xa)
			  printf("10FD\n");
			else if (dat == 0xb)
			  printf("100FD\n");
		}

		x = vtu_readwait(VTU_OPS_REGISTER);
		phy_write(GLOBAL_REGS_1, VTU_VID_REG, 0xFFF);
		while(1) {
			phy_write(GLOBAL_REGS_1, VTU_OPS_REGISTER, (1 << 15) | VTU_OP_GET_NEXT);
			vtu_readwait(VTU_OPS_REGISTER);
			phy_read(GLOBAL_REGS_1, VTU_VID_REG, &x);

			// No more VTU entries
			if ((x & 0xFFF) == 0xFFF) {
				printf("vtu_total=%d\n", vtu);
				break;
			}
			vtu++;
			printf("vtu%d_vid=%d\n", vtu, x & 0xfff);
			for (port = 0; port < 7; port++) {
				phy_read(0x18 + port, 7, &r7);
				if(port == 0)
				  phy_read(GLOBAL_REGS_1, 0x07, &x);
				else if (port == 4)
				  phy_read(GLOBAL_REGS_1, 0x08, &x);


				switch(port) {
				  case 0:
					if ((x & 0x0003) != 0x0003) {
						printf("vtu%d_port0=1\n"
						  "vtu%d_port0_forcevid=%c\n"
						  "vtu%d_port0_egress=%s\n"
						  "vtu%d_port0_alias=%s\n",
						  vtu,
						  vtu, r7 & 0x1000? 'Y':'N',
						  vtu, MemberTagString(x & 3),
						  vtu, !opt_ethbb ? "a":"cpu");
					}
					break;
				  case 1:
					if ((x & 0x0030) != 0x0030) {
						printf("vtu%d_port1=1\n"
						  "vtu%d_port1_forcevid=%c\n"
						  "vtu%d_port1_egress=%s\n"
						  "vtu%d_port1_alias=%s\n",
						  vtu,
						  vtu, r7 & 0x1000? 'Y':'N',
						  vtu, MemberTagString((x>>4)&3),
						  vtu, !opt_ethbb ? "b":"a");
					}
					break;
				  case 2:
					if ((x & 0x0300) != 0x0300) {
						printf("vtu%d_port2=1\n"
						  "vtu%d_port2_forcevid=%c\n"
						  "vtu%d_port2_egress=%s\n"
						  "vtu%d_port2_alias=%s\n",
						  vtu,
						  vtu, r7 & 0x1000? 'Y':'N',
						  vtu, MemberTagString((x>>8)&3),
						  vtu, !opt_ethbb ? "unused":"b");
					}
					break;
				  case 3:
					if ((x & 0x3000) != 0x3000) {
						printf("vtu%d_port3=1\n"
						  "vtu%d_port3_forcevid=%c\n"
						  "vtu%d_port3_egress=%s\n"
						  "vtu%d_port3_alias=%s\n",
						  vtu,
						  vtu, r7 & 0x1000? 'Y':'N',
						  vtu, MemberTagString((x>>12)&3),
						  vtu, !opt_ethbb ? "unused":"c");
					}
					break;
				  case 4:
					if ((x & 0x0003) != 0x0003) {
						printf("vtu%d_port4=1\n"
						  "vtu%d_port4_forcevid=%c\n"
						  "vtu%d_port4_egress=%s\n"
						  "vtu%d_port4_alias=%s\n",
						  vtu,
						  vtu, r7 & 0x1000? 'Y':'N',
						  vtu, MemberTagString(x & 3),
						  vtu, !opt_ethbb ? "b":"d");
					}
					break;
				  case 5:
					if ((x & 0x0030) != 0x0030) {
						printf("vtu%d_port5=1\n"
						  "vtu%d_port5_forcevid=%c\n"
						  "vtu%d_port5_egress=%s\n"
						  "vtu%d_port5_alias=%s\n",
						  vtu,
						  vtu, r7 & 0x1000? 'Y':'N',
						  vtu, MemberTagString((x>>4)&3),
						  vtu, !opt_ethbb ? "cpu":"unused");
					}
					break;
				  case 6:
					if ((x & 0x0300) != 0x0300) {
						printf("vtu%d_port6=1\n"
						  "vtu%d_port6_forcevid=%c\n"
						  "vtu%d_port6_egress=%s\n"
						  "vtu%d_port6_alias=%s\n",
						  vtu,
						  vtu, r7 & 0x1000? 'Y':'N',
						  vtu, MemberTagString(x & 3),
						  vtu, "unused");
					}
					break;
				}
			}
		}
		nbuslock();
	}

	if(opt_greenledon && !is_7400()) poke16(0x2, (peek16(0x2) | 0x8000));
	if(opt_greenledoff && !is_7400()) poke16(0x2, (peek16(0x2) & ~0x8000));
	if((opt_greenledon || opt_greenledoff) && is_7400()) {
		if(model == 0x7680) {
			/* Green LED is 0_5 */
			mxgpioregs[0x104/4] = (0x3 << 10);
			if(opt_greenledon)
			  mxgpioregs[0x708/4] = (1 << 5);
			if(opt_greenledoff)
			  mxgpioregs[0x704/4] = (1 << 5);
			mxgpioregs[0xB04/4] = (1 << 5);
		} else {
			/* Green LED is 0_28 */
			mxgpioregs[(0x114)/4] = (0x3 << 24); 
			if(opt_greenledon)
			  mxgpioregs[(0x708)/4] = (0x1 << 28);
			if(opt_greenledoff)
			  mxgpioregs[(0x704)/4] = (0x1 << 28);
			mxgpioregs[(0xB04)/4] = (0x1 << 28);
		}
	}

	if(opt_redledon && !is_7400()) poke16(0x2, (peek16(0x2) | 0x4000));
	if(opt_redledoff && !is_7400()) poke16(0x2, (peek16(0x2) & ~0x4000));
	if((opt_redledon || opt_redledoff) && is_7400()) {
		if(model == 0x7680) {
			/* Red LES is 0_7 */
			mxgpioregs[0x104/4] = (0x3 << 14);
			if(opt_redledon)
			  mxgpioregs[0x708/4] = (1 << 7);
			if(opt_redledoff)
			  mxgpioregs[0x704/4] = (1 << 7);
			mxgpioregs[0xB04/4] = (1 << 7);
		} else {
			/* Red LED is 0_17 */
			mxgpioregs[(0x114)/4] = (0x3 << 2); 
			if(opt_redledon)
			  mxgpioregs[(0x708)/4] = (0x1 << 17);
			if(opt_redledoff)
			  mxgpioregs[(0x704)/4] = (0x1 << 17);
			mxgpioregs[(0xB04)/4] = (0x1 << 17);
		}
	}

	if(opt_txenon && !is_7400()) {
		int j = 0, sz = strlen(txen_on);
		char *st = 0, *en = 0;
		for(st = txen_on; j < sz; st = en + 1) {
			int xu = strtoul(st, &en, 0);
			j += (en - st) + 1;
			if(xu >= 0 && xu <= 7)
			  poke16(0x2e, (peek16(0x2e) | (1 << xu)));
		}
	}
	if(opt_txenoff && !is_7400()) {
		int j = 0, sz = strlen(txen_off);
		char *st = 0, *en = 0;
		for(st = txen_off; j < sz; st = en + 1) {
			int xu = strtoul(st, &en, 0);
			j += (en - st) + 1;
			if(xu >= 0 && xu <= 7)
			  poke16(0x2e, (peek16(0x2e) & ~(1 << xu)));
		}
	}

	if(opt_bbclkoff && model == 0x4600)
	  poke16(0x2e, (peek16(0x2e) & ~(0x300)));
	if(opt_bbclkon && model == 0x4600)
	  poke16(0x2e, (peek16(0x2e) | (0x100)));
	if(opt_bbclk2on && model == 0x4600)
	  poke16(0x2e, (peek16(0x2e) | (0x300)));
	if(opt_bbclk3on && model == 0x4600) poke16(0x2e, (peek16(0x2e) | (0x200)));

	if(opt_485speed && model == 0x7670) {
		unsigned int *mxpwmregs = NULL;
		unsigned short speed;
		mxpwmregs = mmap(0, getpagesize(), PROT_READ|PROT_WRITE,
		  MAP_SHARED, devmem, 0x80064000);
		//3_18
		mxgpioregs[0x178/4] = 0x3 << 4;

		speed = (24000000/((opt_485speed*1.275)-1));
		mxpwmregs[0x8/4] = 0x3<<30;
		mxpwmregs[0x4/4] = 1<<2;
		mxpwmregs[0x50/4] = (((speed/2)<<16));
		mxpwmregs[0x60/4] = (0x2<<18) | (0x3<<16) | speed;
	}

	if(opt_modbuspoweron && (model == 0x7670 || model == 0x7680)) {
		mxgpioregs[0x718/4] = 1 << 15;
		mxgpioregs[0xb14/4] = 1 << 15;
		mxgpioregs[0xb18/4] = 1 << 14;
		usleep(10000);
		if(((mxgpioregs[0x910/4] >> 14) & 0x1)) {
			mxgpioregs[0x714/4] = 1 << 15;
			printf("modbuspoweron=0\n");
		} else {
			mxgpioregs[0x714/4] = 1 << 15;
			mxgpioregs[0x714/4] = 1 << 13;
			mxgpioregs[0xb14/4] = 1 << 13;
			printf("modbuspoweron=1\n");
		}
	}
	if(opt_modbuspoweroff && (model == 0x7670 || model == 0x7680)) {
		mxgpioregs[0x718/4] = (1 << 13);
		mxgpioregs[0x714/4] = (1 << 15);
	}

	if(opt_pokef && model == 0x7680) {
		fpga_write(opt_address, &opt_pokeval, 1);
	}

	if(opt_peekf && model == 0x7680) {
		unsigned char buf;
		fpga_read(opt_address, &buf, 1);
		nbusunlock();
		printf("0x%x\n", buf);
		nbuslock();
	}

	if(opt_485_0speed && model == 0x7680) {
		char buf[3];
		unsigned int tempvar;
		tempvar = (50000000/(opt_485_0speed/9.5));
		buf[0] = ((tempvar >> 16) & 0xFF);
		buf[1] = ((tempvar >> 8) & 0xFF);
		buf[2] = (tempvar & 0xFF);
		fpga_write(32, &buf, 3);

		tempvar = (50000000/(opt_485_0speed/0.5));
		buf[0] = ((tempvar >> 16) & 0xFF);
		buf[1] = ((tempvar >> 8) & 0xFF);
		buf[2] = (tempvar & 0xFF);
		fpga_write(35, &buf, 3);
	}
	if(opt_485_1speed && model == 0x7680) {
		char buf[3];
		unsigned int tempvar;
		tempvar = (25000000/(opt_485_1speed/9.5));
		buf[0] = ((tempvar >> 16) & 0xFF);
		buf[1] = ((tempvar >> 8) & 0xFF);
		buf[2] = (tempvar & 0xFF);
		fpga_write(38, &buf, 3);

		tempvar = (25000000/(opt_485_1speed/0.5));
		buf[0] = ((tempvar >> 16) & 0xFF);
		buf[1] = ((tempvar >> 8) & 0xFF);
		buf[2] = (tempvar & 0xFF);
		fpga_write(41, &buf, 3);
	}

	//NOTE: opt_dacX is << by 1
	if(opt_dac0 && model == 0x7680) {
		char buf[2];
		buf[0] = ((opt_dac0 >> 9) & 0xf);
		buf[1] = ((opt_dac0 >> 1) & 0xff);
		fpga_write(23, &buf, 2);
	}
	if(opt_dac1 && model == 0x7680) {
		char buf[2];
		buf[0] = ((opt_dac1 >> 9) & 0xf);
		buf[1] = ((opt_dac1 >> 1) & 0xff);
		fpga_write(25, &buf, 2);
	}
	if(opt_dac2 && model == 0x7680) {
		char buf[2];
		buf[0] = ((opt_dac2 >> 9) & 0xf);
		buf[1] = ((opt_dac2 >> 1) & 0xff);
		fpga_write(27, &buf, 2);
	}
	if(opt_dac3 && model == 0x7680) {
		char buf[2];
		buf[0] = ((opt_dac3 >> 9) & 0xf);
		buf[1] = ((opt_dac3 >> 1) & 0xff);
		fpga_write(29, &buf, 2);
	}




	/* Should be last! */
	if (opt_watchdog) {
		struct sched_param sched;
		int fd, r, grain = 0, mode, t;
		char c, time = 3;
		unsigned char dat[3] = {0x0};
		unsigned short x = 0;
		fd_set rfds;
		struct timeval tv_now, tv_wait, tv_exp;
		nbusunlock();

#define FEED0 338000
#define FEED1 2706000
#define FEED2 10824000
		/* Single character writes to /dev/watchdog keep it fed */
		/* XXX: DO NOT RUN MULTIPLE INSTANCES OF THIS! IT WILL 
		 * RESULT IN UNDEFINED BEHAVIOR!
		 */
		/* New /dev/watchdog interface:
		 * - Still supports the standard interface that expects
		 *   '0' '1' '2' or '3'.
		 * - Most other characters are still treated as 3.
		 * - 'f' (for feed) is a special character used to create
		 *   automatic feeds.
		 * - 'f' is expected to be followed by three digits, eg "f100".
		 * - Those digits are interpreted as a 3 digit number, which
		 *   is then divided by 10.  For example, "f456" feeds the
		 *   watchdog for 45.6 seconds.
		 * - 'a' (for autofeed) is a special character used to 
		 *   create an autofeed situation
		 * - In order to guarantee timeout, 'a' is expected to be
		 *   followed by a single number, 0 through 3, eg "a1"
		 * - This single number is the standard WDT interface
		 *   and is not real world time
		 * - An extended feed is canceled by any new write to
		 *   /dev/watchdog.
		 * - If 'f' is not followed by 3 digits, the behavior is
		 *   undefined, but in practice you typically just fed the
		 *   wdt for a long time.
		 * - If 'a' is not followed by a single digit, the behavior 
		 *   is undefined.
		 */

		/* We need to make sure we create the pipe before daemonizing
		 * so that as soon as the call to tshwctl -W completes we can
		 * write to the pipe.  The write side of the pipe NEEDS to be 
		 * opened with O_WRONLY and !O_NONBLOCK in order to ensure
		 * that any immediate writes will block until we open the 
		 * pipe here 
		 *
		 * bash redirects open the file with O_WRONLY and !O_NONBLOCK
		 */
		mknod("/dev/watchdog", S_IFIFO|0666, 0);
		daemon(0, 0);
		sched.sched_priority = 98;
		sched_setscheduler(0, SCHED_FIFO, &sched);
		mode = 0; // mode 1 == processing a multi-feed write
		while(1) {
			fd = open("/dev/watchdog", O_RDONLY | O_NONBLOCK);
			assert (fd != -1);
			FD_ZERO(&rfds);
			FD_SET(fd, &rfds);
			if (mode)
				select(fd + 1, &rfds, NULL, NULL, &tv_wait);
			else
				select(fd + 1, &rfds, NULL, NULL, NULL);
			r = 0;
			if (FD_ISSET(fd, &rfds)) r = read(fd, &c, 1);
			if (r) { // data coming in /dev/watchdog
				if (c == 'f') { // new style feed
					read(fd, &c, 1);
					t = (c - '0') * 100;
					read(fd, &c, 1);
					t += (c - '0') * 10;
					read(fd, &c, 1);
					t += (c - '0');
					if(is_7400()) {
						dat[1] = ((t >> 8) & 0xff);
						dat[2] = (t & 0xff);
						nbuslock();
						m0_write(&dat, 3);
						nbusunlock();
						mode = 0;
					} else {
						t = t * 100000;
						x = 0;
						if (t > FEED1) x = 1;
						if (t > FEED2) x = 2;
						nbuslock();
						poke16(0x2a, x);
						nbusunlock();
						switch (x) {
							case 0: grain = FEED0; break;
							case 1: grain = FEED1; break;
							case 2: grain = FEED2; break;
						}
						mode = 1;
						gettimeofday(&tv_exp, NULL);
						t += tv_exp.tv_usec;
						tv_exp.tv_usec = t % 1000000;
						tv_exp.tv_sec += t / 1000000;
					}
				/*XXX: This is needed because close() is not
				 * guaranteed to have closed and flushed the
				 * pipe by the time we re-open and read again
				 */ 
				} else if(c == '\n') { 
				} else { // old style feed
					mode = 0;
					if(c == 'a') {
						mode = 2;
						read(fd, &c, 1);
					}
					if (c > '3' || c < '0') c = '3';
					nbuslock();
					if(is_7400()) {
						dat[1] = 0;
						dat[2] = c - '0';
						m0_write(&dat, 3);
					} else poke16(0x2a, c - '0');
					nbusunlock();
					if(c == '3') mode = 0;
					/* See XXX: above */
					time = (c - '0');
				}
			}
			if (mode == 1) {
				// if in mode 1, we feed wdt and calculate
				// how long to sleep
				nbuslock();
				poke16(0x2a, x);
				nbusunlock();
				gettimeofday(&tv_now, NULL);
				t = tv_exp.tv_sec - tv_now.tv_sec;
				t *= 1000000;
				t += tv_exp.tv_usec;
				t -= tv_now.tv_usec;
				if (t <= grain) mode = 0;
				else if (t <= 3*grain/2) t -= grain;
				else t = grain/2;
				tv_wait.tv_sec = t / 1000000;
				tv_wait.tv_usec = t % 1000000;
			}
			if(mode == 2) {
				nbuslock();
				if(is_7400()) {
					dat[1] = 0;
					dat[2] = time;
					m0_write(&dat, 3);
				} else poke16(0x2a, time);
				nbusunlock();
				switch (time) {
					case 0: 
					  tv_wait.tv_sec = (FEED0 / 2) / 1000000;
					  tv_wait.tv_usec = (FEED0 / 2) % 1000000;
					  break;
					case 1:
					  tv_wait.tv_sec = (FEED1 / 2) / 1000000;
					  tv_wait.tv_usec = (FEED1 / 2) % 1000000;
					  break;
					case 2:
					  tv_wait.tv_sec = (FEED2 / 2) / 1000000;
					  tv_wait.tv_usec = (FEED2 / 2) % 1000000;
					  break;
				}
			}
			// if mode was 0 and no character was read, don't do
			// anything, just go back to waiting for a character
			close(fd);
		} // end forever loop
	}

	nbusunlock();
	return 0;
}
